import keyword
import sys

SIMMODEL_KEYWORDS = keyword.kwlist + [
    # BasicRtlSimModel properties
    'sim',
    '_interfaces',
    '_processes',
    '_units',
    '_outputs',
    '_init_body',
    '__init__',
    '__new__',
    '__getattr__',
    '__getattribute__',
    '__setattr__',
    '__setattribute__',
    # imports
    'pyMathBitPrecise',
    'hwtSimApi',
    'Array3t',
    'Array3val',
    'Bits3t',
    'Bits3val',
    'define_Enum3t',
    'BasicRtlSimModel',
    'BasicRtlSimProxy',
    'sensitivity',
    'connectSimPort',
    'sim_eval_cond',
    'slice',
]

if sys.version_info[0] <= 2:
    SIMMODEL_KEYWORDS.append("None")



from hdlConvertorAst.hdlAst._statements import HdlStmIf, HdlStmBlock


def elifs_to_if_then_else(stm):
    """
    Optionally create if-then-else without else-ifs from this if-then-else statement

    :type stm: HdlStmIf
    :note: non recursive
    """
    if stm.elifs:
        # replace elifs with nested if statements
        ifFalse = HdlStmBlock()
        topIf = HdlStmIf(stm.cond, stm.if_true, ifFalse)

        for c, stms in stm.elifs:
            _ifFalse = HdlStmBlock()

            lastIf = HdlStmIf(c, stms, _ifFalse)

            ifFalse.append(lastIf)
            ifFalse = _ifFalse

        if stm.if_false is None:
            lastIf.if_false = HdlStmBlock()
        else:
            lastIf.if_false = stm.if_false

        return topIf
    return stm



"""

This package contains the classes which do convert hdlConvertorAst.hdlAst to a target language (e.g. Verilog/SystemVerilog/VHDL)

:note: The convertors in this package do not modify the AST itself, if the input comes from a different language.
    It may be required to translate the AST first using e.g. :mod:`hdlConvertorAst.translate`.

.. uml:: hdlConvertorAst.to
    :classes:
    :packages:

"""


from hdlConvertorAst.hdlAst import HdlValueId, HdlValueInt, HdlOp, HdlOpType
from hdlConvertorAst.py_ver_compatibility import is_str
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_getattr, \
    hdl_name_prefix


def add_io_prefix(o):
    s = HdlValueId("self")
    s = hdl_getattr(s, "io")
    return hdl_name_prefix(s, o)


class ApplyIoScopeToSignalNames(HdlAstVisitor):
    """
    Change every signal_name to self.io.signal_name
    in every statement expression.
    And for signal which are beeing read add .val.

    :ivar ~._stm_dst: flag which tells if currently processing destination of HdlStmAssig
    """

    def __init__(self):
        super(ApplyIoScopeToSignalNames, self).__init__()
        self._stm_dst = False

    def visit_HdlIdDef(self, var):
        """
        :type var: HdlIdDef
        """
        pass

    def visit_port(self, o):
        pass

    def visit_param(self, o):
        pass

    def visit_HdlCompInst(self, o):
        pass

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        :return: iHdlExpr
        """
        if o.fn == HdlOpType.DOT:
            # update only left most name
            o.ops[0] = self.visit_iHdlExpr(o.ops[0])
            if not self._stm_dst:
                return hdl_getattr(o, "val")
        else:
            o.ops = [self.visit_iHdlExpr(op) for op in o.ops]

        return o

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        :return: iHdlExpr
        """
        if isinstance(o, HdlValueId):
            o = add_io_prefix(o)
            if not self._stm_dst:
                return hdl_getattr(o, "val")
            return o
        elif o is None or isinstance(o, HdlValueInt) or is_str(o):
            return o
        elif isinstance(o, list):
            return [self.visit_iHdlExpr(_o) for _o in o]
        elif isinstance(o, HdlOp):
            return self.visit_HdlOp(o)
        else:
            raise NotImplementedError(
                "Do not know how to convert %s" % (o))

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        assert o.event_delay is None, o
        assert o.time_delay is None, o

        _stm_dst = self._stm_dst
        self._stm_dst = True
        try:
            o.dst = self.visit_iHdlExpr(o.dst)
        finally:
            self._stm_dst = _stm_dst

        o.src = self.visit_iHdlExpr(o.src)

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        o.cond = self.visit_iHdlExpr(o.cond)
        if o.if_true is not None:
            self.visit_iHdlStatement(o.if_true)
        o.elifs = [
            (self.visit_iHdlExpr(c), stm)
            for (c, stm) in o.elifs
        ]
        for _, stm in o.elifs:
            self.visit_iHdlStatement(stm)
        if o.if_false is not None:
            self.visit_iHdlStatement(o.if_false)



"""
This package contains the containers
for Hardware Description Language Abstract Syntax Tree (HDL AST)

(System) Verilog/VHDL can be converted by this format and back using this library
Main container is HdlContext which can store all other objects.

:note: this module is using PEP484 type hints, and the type hints are in comments
    to sustain Python 2.7 compatibility
:attention: this module is just simple format of HDL and does not performs all check,
    e.g. you can use a string as a type of the port,
    but such a module generally can not be synthetised to a hardware

:attention: classes in his module are used from C++


.. uml:: hdlConvertorAst.hdlAst
    :classes:
    :packages:
"""

from hdlConvertorAst.hdlAst._bases import (
    iHdlObj, iHdlObjInModule, iHdlObjWithName, iHdlStatement)
from hdlConvertorAst.hdlAst._defs import HdlFunctionDef, HdlIdDef
from hdlConvertorAst.hdlAst._expr import (
    HdlAll, HdlOpType, HdlOp, HdlDirection, HdlValueInt, HdlValueId,
    HdlOthers, HdlTypeAuto, HdlTypeType, HdlTypeSubtype, HdlExprNotImplemented,
    iHdlExpr)
from hdlConvertorAst.hdlAst._statements import (
    HdlImport, HdlStmAssign, HdlStmBlockJoinType, HdlStmBlock, HdlStmBreak,
    HdlStmCaseUniqConstrain, HdlStmCaseType, HdlStmCase, HdlStmContinue, HdlStmFor, HdlStmForIn, HdlStmIf,
    HdlStmProcessTriggerConstrain, HdlStmProcess, HdlStmRepeat, HdlStmReturn, HdlStmWait,
    HdlStmWhile, HdlStmThrow, HdlStmNop)
from hdlConvertorAst.hdlAst._structural import (
    HdlCompInst, HdlContext, HdlModuleDec, HdlModuleDef, HdlValueIdspace,
    HdlLibrary)
from hdlConvertorAst.hdlAst._typeDefs import (
    iHdlTypeDef, HdlClassType, HdlClassDef, HdlPhysicalDef, HdlEnumDef, HdlTypeBitsDef)
from hdlConvertorAst.hdlAst.utils import CodePosition
from hdlConvertorAst.hdlAst._statements import ALL_STATEMENT_CLASSES


NON_INSTANCIABLE_NODES = {
    HdlAll,
    HdlOthers,
    HdlTypeType,
    HdlTypeSubtype,
    HdlTypeAuto,
}


from hdlConvertorAst.hdlAst import HdlOpType, HdlValueId, HdlValueInt, \
    HdlOp
from hdlConvertorAst.py_ver_compatibility import is_str
from hdlConvertorAst.to.common import ToHdlCommon, ASSOCIATIVITY
from hdlConvertorAst.to.hdlUtils import iter_with_last, Indent


L = ASSOCIATIVITY.L_TO_R
R = ASSOCIATIVITY.R_TO_L


class ToBasicHdlSimModelExpr(ToHdlCommon):
    OP_PRECEDENCE = {
        # TO/DOWNTO becomes a call to slice
        # HdlOpType.DOWNTO: 1,
        # HdlOpType.TO: 1,

        # note that HdlExpressions in BasicHdlSimModel
        # do not use == but ._eq()
        HdlOpType.EQ: (11, L),
        HdlOpType.NE: (11, L),
        HdlOpType.GT:  (11, L),
        HdlOpType.LT:  (11, L),
        HdlOpType.GE:  (11, L),
        HdlOpType.LE:  (11, L),
        HdlOpType.IS:  (11, L),
        HdlOpType.IS_NOT:  (11, L),

        HdlOpType.OR: (10, L),
        HdlOpType.XOR: (10, L),
        HdlOpType.AND: (9, L),

        HdlOpType.SLL: (8, L),
        HdlOpType.SRL: (8, L),

        HdlOpType.ADD: (7, L),
        HdlOpType.SUB: (7, L),

        HdlOpType.DIV: (6, L),
        HdlOpType.MUL: (6, L),
        HdlOpType.MOD: (6, L),

        HdlOpType.NEG_LOG: (5, R),
        HdlOpType.NEG: (5, R),
        HdlOpType.MINUS_UNARY: (5, R),
        HdlOpType.POW: (4, R),

        HdlOpType.INDEX: (1, L),

        HdlOpType.RISING: (1, L),
        HdlOpType.FALLING: (1, L),

        # concat/ternary become a call to _concat, _ternary__val function
        # HdlOpType.CONCAT: 2,
        # HdlOpType.TERNARY: 2,
        # rising/faling as ._onRisingEdge(), ._onFallingEdge()
        HdlOpType.CALL: (1, L),
        # parametrization values are parameters of component class
        # constructor
        HdlOpType.PARAMETRIZATION: (1, L),

        HdlOpType.DOT: (1, L),
    }
    _unaryEventOps = {
        HdlOpType.RISING: "._onRisingEdge()",
        HdlOpType.FALLING: "._onFallingEdge()",
    }
    GENERIC_UNARY_OPS = {
        HdlOpType.NEG_LOG: "not ",
        HdlOpType.NEG: "~",
        HdlOpType.MINUS_UNARY: "-",
    }

    GENERIC_BIN_OPS = {
        HdlOpType.AND: " & ",
        HdlOpType.OR: " | ",
        HdlOpType.XOR: " ^ ",

        HdlOpType.EQ: ' == ',
        HdlOpType.NE: " != ",

        HdlOpType.IS: ' is ',
        HdlOpType.IS_NOT: " is not ",

        HdlOpType.MUL: " * ",
        HdlOpType.DIV: " // ",
        HdlOpType.POW: " ** ",
        HdlOpType.MOD: " % ",

        HdlOpType.SLL: " << ",
        HdlOpType.SRL: " >> ",
    }
    GENERIC_BIN_OPS.update(ToHdlCommon.GENERIC_BIN_OPS)

    def visit_HdlValueInt(self, o):
        """
        :type o: HdlValueInt
        """
        w = self.out.write
        if o.bits is None:
            w(str(o.val))
        else:
            if o.base is None:
                w(str(o.val))
            else:
                b = o.base
                v = o.val
                if is_str(v):
                    if set(v) == set('x'):
                        v = None
                        f = "{0}"
                    elif b == 2:
                        f = "0b{0}"
                    elif b == 8:
                        f = "0o{0}"
                    elif b == 10:
                        f = "{0}"
                    elif b == 16:
                        f = "0x{0}"
                    else:
                        raise NotImplementedError(b)
                else:
                    if b == 2:
                        f = "0b{0:b}"
                    elif b == 8:
                        f = "0o{0:o}"
                    elif b == 10:
                        f = "{0:d}"
                    elif b == 16:
                        f = "0x{0:x}"
                    else:
                        raise NotImplementedError(b)
                v = f.format(v)
                w(v)

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        """
        w = self.out.write
        if isinstance(o, HdlValueId):
            w(o.val)
            return
        elif is_str(o):
            w('"%s"' % o)
            return
        elif isinstance(o, HdlValueInt):
            self.visit_HdlValueInt(o)
            return
        elif isinstance(o, (list, tuple)):
            with_nl = len(o) > 3
            w("(")
            for elem in o:
                self.visit_iHdlExpr(elem)
                if with_nl:
                    w(", \n")
                else:
                    w(", ")
            w(")")
            return
        elif isinstance(o, HdlOp):
            self.visit_HdlOp(o)
            return
        elif o is None:
            w("None")
            return
        elif isinstance(o, dict):
            w("{")
            with Indent(self.out):
                for last, (k, v) in iter_with_last(sorted(o.items(), key=lambda x: x[0])):
                    self.visit_iHdlExpr(k)
                    w(": ")
                    self.visit_iHdlExpr(v)
                    if not last:
                        w(",\n")
                    else:
                        w("\n")
            w("}")
            return
        elif isinstance(o, float):
            w("%f" % o)
            return

        raise NotImplementedError(o.__class__, o)

    def visit_HdlOp(self, o):
        """
        :type op: HdlOp
        """
        ops = o.ops
        op = o.fn
        w = self.out.write

        op_str = self._unaryEventOps.get(op, None)
        if op_str is not None:
            op0 = ops[0]
            w(op0.val)
            w(op_str)
        elif op == HdlOpType.MAP_ASSOCIATION:
            # kwargs, [todo]: dict constructor
            self._visit_operand(o.ops[0], 0, o, False, False)
            w("=")
            self._visit_operand(o.ops[1], 1, o, False, False)
        elif op == HdlOpType.TERNARY:
            # not that this is not a ternary in hdl epxression but in python wich is evaluated
            # during object construction, the hdl ternary is _ternary function
            c, o1, o2 = o.ops
            self._visit_operand(o1, 0, o, False, False)
            w(" if ")
            self._visit_operand(c, 0, o, False, False)
            w(" else ")
            self._visit_operand(o2, 0, o, False, False)
        else:
            return super(ToBasicHdlSimModelExpr, self).visit_HdlOp(o)



from hdlConvertorAst.hdlAst import HdlOpType
from hdlConvertorAst.to.hdl_ast_modifier import HdlAstModifier
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call
from hdlConvertorAst.to.verilog.expr import ASSIGN_OPERATORS


class SignalAssignmentsToCallOp(HdlAstModifier):
    """
    Convert an assignment to an call operator.
    """

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        assert o.event_delay is None, o
        assert o.time_delay is None, o
        return hdl_call(o.dst, [o.src, ])

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        :return: HdlOp
        """
        if o.fn == HdlOpType.ASSIGN:
            dst, src = o.ops
            return hdl_call(dst, [src, ])
        elif o.fn in ASSIGN_OPERATORS:
            raise NotImplementedError(o)

        return o



from hdlConvertorAst.hdlAst._expr import HdlValueInt


class Indent(object):
    """
    indentation context
    """

    def __init__(self, autoIndentStream):
        self.s = autoIndentStream
        self.original_indent = None

    def __enter__(self):
        s = self.s
        self.original_indent = s.indent_str
        s.indent_cnt += 1
        s.indent_str = s.indent_str + s.INDENT_STEP

    def __exit__(self, exception_type, exception_value, traceback):
        s = self.s
        s.indent_cnt -= 1
        s.indent_str = self.original_indent


class UnIndent():
    """
    unindentation context
    """

    def __init__(self, autoIndentStream):
        self.s = autoIndentStream
        self.original_indent = None

    def __enter__(self):
        s = self.s
        self.original_indent = s.indent_str
        assert s.indent_cnt > 0
        s.indent_cnt -= 1
        s.indent_str = s.indent_str[0:len(s.indent_str) - len(s.INDENT_STEP)]

    def __exit__(self, exception_type, exception_value, traceback):
        s = self.s
        s.indent_cnt += 1
        s.indent_str = self.original_indent


class AutoIndentingStream():

    def __init__(self, stream, indent_step):
        """
        :param stream: output stream
        :param indent_step: string of indent
        """

        self.INDENT_STEP = indent_step
        self.stream = stream
        self.requires_indent = True
        self.indent_cnt = 0
        self.indent_str = ""

    def write(self, s):
        w = self.stream.write
        if self.requires_indent and s != "\n":
            w(self.indent_str)
        w(s)
        self.requires_indent = s.endswith("\n")


def iter_with_last(it):
    # Ensure it's an iterator and get the first field
    it = iter(it)
    try:
        prev = next(it)
    except StopIteration:
        return
    for item in it:
        # Lag by one item so I know I'm not at the end
        yield False, prev
        prev = item

    # Last item
    yield True, prev


def bit_string(v, width, vld_mask=None):
    """
    :type v: int
    :type width: int
    :type vld_mask: Optional[int]
    :param v: integer value of bitstring
    :param widht: number of bits in value
    :param vld_mask: mask which has 1 for every valid bit in value
    :return: HdlValueInt
    """
    all_mask = (1 << width) - 1
    if vld_mask is None:
        vld_mask = all_mask

    # if vld_mask == 0:
    #     if width % 4 == 0:
    #         base = 16
    #         bit_string = "".join(["x" for _ in range(width//4)])
    #     else:
    #         base = 2
    #         bit_string = "".join(["x" for _ in range(width)])

    elif width % 4 == 0 and vld_mask == (1 << width) - 1:
        # hex full valid
        base = 16
        bit_string = ("%0" + str(width // 4) + 'x') % (v)
    else:
        # binary
        base = 2
        buff = []
        for i in range(width - 1, -1, -1):
            mask = (1 << i)
            b = v & mask

            if vld_mask & mask:
                s = "1" if b else "0"
            else:
                s = "x"
            buff.append(s)
        bit_string = "".join(buff)
    return HdlValueInt(bit_string, width, base)



from hdlConvertorAst.hdlAst import HdlDirection, iHdlStatement, \
    HdlIdDef, HdlModuleDec, HdlFunctionDef, HdlCompInst, \
    HdlTypeType, HdlOp, HdlOpType, HdlValueIdspace, HdlPhysicalDef, \
    HdlEnumDef, HdlTypeSubtype, HdlValueInt, HdlClassDef, HdlClassType, \
    HdlValueId
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last, UnIndent
from hdlConvertorAst.to.vhdl.stm import ToVhdl2008Stm


class ToVhdl2008(ToVhdl2008Stm):
    """
    Convert hdlObject AST back to VHDL
    """
    DIR2V = {
        HdlDirection.IN: "IN",
        HdlDirection.OUT: "OUT",
        HdlDirection.INOUT: "INOUT",
    }

    def __init__(self, out_stream):
        ToVhdl2008Stm.__init__(self, out_stream)
        self.in_typedef = False

    def visit_doc(self, obj):
        return super(ToVhdl2008, self).visit_doc(obj, "--")

    def visit_direction(self, d):
        vd = self.DIR2V[d]
        self.out.write(vd)

    def visit_main_obj(self, o):
        ToVhdl2008Stm.visit_main_obj(self, o)
        add_nl = isinstance(o, (HdlModuleDec, HdlModuleDec, HdlValueIdspace))
        if add_nl:
            self.out.write("\n")

    def visit_param_or_port_declr(self, o, is_param):
        """
        :type p: HdlIdDef
        """
        self.visit_doc(o)
        w = self.out.write
        w(o.name)
        w(" : ")
        if not is_param:
            d = o.direction
            if d != HdlDirection.INTERNAL:
                self.visit_direction(d)
                w(" ")
        self.visit_type(o.type)
        v = o.value
        if v is not None:
            w(" := ")
            self.visit_iHdlExpr(v)

    def visit_HdlModuleDec(self, e, vhdl_obj_name="ENTITY"):
        """
        :param e: Entity
        :type e: HdlModuleDec
        """
        self.visit_doc(e)
        w = self.out.write
        w(vhdl_obj_name)
        w(" ")
        w(e.name)
        w(" IS\n")
        gs = e.params
        if gs:
            with Indent(self.out):
                w("GENERIC(\n")
                with Indent(self.out):
                    for last, g in iter_with_last(gs):
                        self.visit_param_or_port_declr(g, True)
                        if last:
                            w("\n")
                        else:
                            w(";\n")

                w(");\n")
        ps = e.ports
        if ps:
            with Indent(self.out):
                w("PORT(\n")
                with Indent(self.out):
                    for last, p in iter_with_last(ps):
                        self.visit_param_or_port_declr(p, False)
                        if last:
                            w("\n")
                        else:
                            w(";\n")
                w(");\n")
        di = e.objs
        if di:
            with Indent(self.out):
                for o in di:
                    self.visit_iHdlObj(o)
        w("END ")
        w(vhdl_obj_name)
        w(";\n")

    def visit_component(self, o):
        """
        :type o: HdlModuleDec
        """
        self.visit_HdlModuleDec(o, vhdl_obj_name="COMPONENT")

    def visit_type(self, t):
        """
        :type t: iHdlExpr
        """
        self.visit_iHdlExpr(t)

    def visit_map_item(self, item):
        self.visit_iHdlExpr(item)

    def visit_map(self, map_):
        w = self.out.write
        with Indent(self.out):
            for last, m in iter_with_last(map_):
                self.visit_map_item(m)
                if last:
                    w("\n")
                else:
                    w(",\n")

    def visit_HdlCompInst(self, c):
        """
        :type c: HdlCompInst
        """
        self.visit_doc(c)
        w = self.out.write
        w(c.name.val)
        w(": ")
        if not isinstance(c.module_name, HdlValueId):
            w("ENTITY ")
        self.visit_iHdlExpr(c.module_name)
        gms = c.param_map
        if gms:
            w(" GENERIC MAP(\n")
            self.visit_map(gms)
            w(")")

        pms = c.port_map
        if pms:
            w(" PORT MAP(\n")
            self.visit_map(pms)
            w(")")
        w(";")

    def visit_body_items(self, objs):
        w = self.out.write
        in_def_section = True
        with Indent(self.out):
            for o in objs:
                if isinstance(o, HdlIdDef):
                    assert in_def_section, o
                    self.visit_HdlIdDef(o)
                    continue
                elif isinstance(o, HdlModuleDec):
                    assert in_def_section, o
                    self.visit_component(o)
                    continue
                elif isinstance(o, HdlFunctionDef):
                    assert in_def_section, o
                    self.visit_HdlFunctionDef(o)
                    continue

                if in_def_section:
                    with UnIndent(self.out):
                        w("BEGIN\n")
                    in_def_section = False

                if isinstance(o, HdlCompInst):
                    self.visit_HdlCompInst(o)
                    w("\n")
                elif isinstance(o, iHdlStatement):
                    self.visit_iHdlStatement(o)
                elif isinstance(o, HdlOp) and o.fn == HdlOpType.CALL:
                    self.visit_HdlOp(o)
                    w(";\n")
                else:
                    raise NotImplementedError(o)
        if in_def_section:
            w("BEGIN\n")

    def visit_HdlModuleDef(self, o):
        """
        :type o: HdlModuleDef
        """
        w = self.out.write
        if o.dec is not None:
            self.visit_HdlModuleDec(o.dec)
            w("\n")
        self.visit_doc(o)
        w("ARCHITECTURE ")
        w(o.name)
        w(" OF ")
        w(o.module_name.val)
        w(" IS\n")
        self.visit_body_items(o.objs)
        self.out.write("END ARCHITECTURE;\n")

    def visit_HdlIdDef(self, var, end=";\n"):
        """
        :type var: HdlIdDef
        """
        self.visit_doc(var)
        w = self.out.write
        name = var.name
        t = var.type
        if t == HdlTypeType:
            assert not var.is_shared
            orig_in_typedef = self.in_typedef
            try:
                self.in_typedef = True
                # typedef
                w("TYPE ")
                w(name)
                w(" IS ")
                _t = var.value
                if isinstance(_t, HdlEnumDef):
                    self.visit_HdlEnumDef(_t)
                elif isinstance(_t, HdlOp):
                    if _t.fn == HdlOpType.INDEX:
                        w("ARRAY (")
                        for last, i in iter_with_last(_t.ops[1:]):
                            self.visit_iHdlExpr(i)
                            if not last:
                                w(", ")
                        w(") OF ")
                        self.visit_iHdlExpr(_t.ops[0])
                    elif _t.fn == HdlOpType.RANGE:
                        w("RANGE ")
                        assert len(_t.ops) == 1, _t.ops
                        self.visit_iHdlExpr(_t.ops[0])
                    else:
                        raise NotImplementedError(_t.fn)

                elif isinstance(_t, HdlClassDef):
                    self.visit_HdlClassDef(_t)
                elif isinstance(_t, HdlPhysicalDef):
                    self.visit_HdlPhysicalDef(_t)
                else:
                    raise NotImplementedError(type(_t))
            finally:
                self.in_typedef = orig_in_typedef
        elif t == HdlTypeSubtype:
            assert not var.is_shared
            orig_in_typedef = self.in_typedef
            try:
                self.in_typedef = True
                w("SUBTYPE ")
                w(name)
                w(" IS ")
                self.visit_iHdlExpr(var.value)
            finally:
                self.in_typedef = orig_in_typedef
        else:
            # signal/variable/port/generic
            if not self.in_typedef:
                latch = var.is_latched
                c = var.is_const
                if c:
                    assert not var.is_shared
                    w("CONSTANT ")
                elif latch:
                    if var.is_shared:
                        w("SHARED ")
                    w("VARIABLE ")
                else:
                    assert not var.is_shared
                    w("SIGNAL ")
            w(name)
            w(" : ")
            self.visit_type(t)
            v = var.value
            if v is not None:
                w(" := ")
                self.visit_iHdlExpr(v)
        w(end)

    def visit_HdlClassDef(self, o):
        """
        :type o: HdlClassDef
        """
        w = self.out.write
        assert o.type == HdlClassType.STRUCT, o.type
        w("RECORD\n")
        with Indent(self.out):
            for m in o.members:
                self.visit_HdlIdDef(m)
        w("END RECORD")

    def visit_HdlPhysicalDef(self, o):
        """
        :type o: HdlPhysicalDef
        """
        w = self.out.write
        self.visit_HdlOp(o.range)
        w("\n")
        with Indent(self.out):
            w("UNITS\n")
            with Indent(self.out):
                for k, v in o.members:
                    w(k)
                    if v is not None:
                        w(" = ")
                        self.visit_HdlOp(v)
                    w(";\n")
            w("END UNITS")

    def visit_HdlEnumDef(self, o):
        """
        :type o: HdlEnumDef
        """
        w = self.out.write
        w('(')
        for last, ev in iter_with_last(o.values):
            k, v = ev
            if k is not None:
                w(k)
            else:
                assert isinstance(v, HdlValueInt) and v.base == 256, v
                self.visit_HdlValueInt(v)

            if not last:
                w(", ")
        w(")")

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        self.visit_doc(o)
        w = self.out.write
        is_procedure = o.return_t is None
        if is_procedure:
            w("PROCEDURE ")
        else:
            w("FUNCTION ")

        w(o.name)
        if o.params:
            w(" (")
            with Indent(self.out):
                for is_last, par in iter_with_last(o.params):
                    self.visit_HdlIdDef(par, end="")
                    if not is_last:
                        w(";\n")
            w(")")
        if not is_procedure:
            w(" RETURN ")
            self.visit_type(o.return_t)
        if o.is_declaration_only:
            w(";\n")
        else:
            w("\n")
            w("IS\n")
            self.visit_body_items(o.body)
            if is_procedure:
                w("END PROCEDURE;\n")
            else:
                w("END FUNCTION;\n")

    def visit_HdlLibrary(self, o):
        """
        :type o: HdlLibrary
        """
        self.visit_doc(o)
        w = self.out.write
        w("LIBRARY ")
        w(o.name)
        w(";\n")

    def visit_HdlImport(self, o):
        """
        :type o: HdlImport
        """
        self.visit_doc(o)
        w = self.out.write
        w("USE ")
        for last, p in iter_with_last(o.path):
            self.visit_iHdlExpr(p)
            if not last:
                w(".")
        w(";\n")

    def visit_HdlValueIdspace(self, o):
        """
        :type o: HdlValueIdspace
        """
        self.visit_doc(o)
        w = self.out.write
        # TODO:: if o.declaration_only:
        w("PACKAGE ")
        w(o.name)
        w(" IS\n")
        with Indent(self.out):
            for _o in o.objs:
                self.visit_main_obj(_o)

        w("END PACKAGE;\n")

    def _visit_HdlStmExitOrNext(self, o, stm_name):
        """
        :type o: HdlOp
        :type stm_name: str

        :attention: next/exit statement is represented as a call of next/exit function
            and does not have a specific object, as it is expression it does not render
            the ending ;
        """
        assert o.fn == HdlOpType.CALL and o.ops[0] == HdlValueId(stm_name), o
        w = self.out.write
        w(stm_name)
        if len(o.ops) == 1:
            pass
        elif len(o.ops) == 2:
            # label
            w(" ")
            self.visit_iHdlExpr(o.ops[1])
        elif len(o.ops) == 3:
            # label, condition
            w(" ")
            _, label, cond = o.ops
            self.visit_iHdlExpr(label)
            w(" WHEN (")
            self.visit_iHdlExpr(cond)
            w(")")
        else:
            raise ValueError(o)

    def visit_HdlOp(self, o):
        arg0 = o.ops[0]
        if o.fn == HdlOpType.CALL and isinstance(arg0, HdlValueId):
            if arg0.val == "EXIT":
                return self.visit_HdlStmExit(o)
            elif arg0.val == "NEXT":
                return self.visit_HdlStmNext(o)

        return ToVhdl2008Stm.visit_HdlOp(self, o)

    def visit_HdlStmExit(self, o):
        """
        :type o: HdlOp
        """
        return self._visit_HdlStmExitOrNext(o, "EXIT")

    def visit_HdlStmNext(self, o):
        """
        :type o: HdlOp
        """
        return self._visit_HdlStmExitOrNext(o, "NEXT")



from hdlConvertorAst.hdlAst import HdlOpType, HdlValueId, HdlValueInt, HdlAll, \
    HdlOp, HdlTypeAuto
from hdlConvertorAst.py_ver_compatibility import is_str
from hdlConvertorAst.to.common import ToHdlCommon, ASSOCIATIVITY, \
    ASSIGN_OPERATORS_SYMBOLS_C
from hdlConvertorAst.to.hdlUtils import iter_with_last
from hdlConvertorAst.to.verilog.utils import collect_array_dims, get_wire_t_params

L = ASSOCIATIVITY.L_TO_R
R = ASSOCIATIVITY.R_TO_L

ASSIGN_OPERATORS = [
    HdlOpType.ASSIGN,
    HdlOpType.PLUS_ASSIGN,
    HdlOpType.MINUS_ASSIGN,
    HdlOpType.MUL_ASSIGN,
    HdlOpType.DIV_ASSIGN,
    HdlOpType.MOD_ASSIGN,
    HdlOpType.AND_ASSIGN,
    HdlOpType.OR_ASSIGN,
    HdlOpType.XOR_ASSIGN,
    HdlOpType.SHIFT_LEFT_ASSIGN,
    HdlOpType.SHIFT_RIGHT_ASSIGN,
    HdlOpType.ARITH_SHIFT_LEFT_ASSIGN,
    HdlOpType.ARITH_SHIFT_RIGHT_ASSIGN,
]

SIGNED = HdlValueId("signed")


def pop_signed_flag(o):
    """
    :type op: HdlOp

    pop signed/unsigned flag from type expr
    """
    base_expr = o
    is_signed = None

    if o.fn == HdlOpType.PARAMETRIZATION and len(o.ops) == 2:
        op1 = o.ops[1]
        if isinstance(op1, HdlOp) and\
                op1.fn == HdlOpType.MAP_ASSOCIATION and\
                op1.ops[0] == SIGNED:
            base_expr = o.ops[0]
            is_signed = bool(int(op1.ops[1]))

    return base_expr, is_signed


class ToVerilog2005Expr(ToHdlCommon):

    GENERIC_BIN_OPS = {
        HdlOpType.APOSTROPHE: "'",

        HdlOpType.AND: " & ",
        HdlOpType.AND_LOG: " && ",
        HdlOpType.OR: " | ",
        HdlOpType.OR_LOG: " || ",
        HdlOpType.NAND: " ~& ",
        HdlOpType.NOR: " ~| ",
        HdlOpType.XOR: " ^ ",
        HdlOpType.XNOR: " ~^ ",

        HdlOpType.DIV: " / ",
        HdlOpType.POW: " ** ",
        HdlOpType.MOD: " % ",

        HdlOpType.EQ: ' == ',
        HdlOpType.NE: " != ",
        HdlOpType.IS: ' === ',
        HdlOpType.IS_NOT: " !== ",

        HdlOpType.EQ_MATCH:" ==? ",
        HdlOpType.NE_MATCH:" !=? ",

        HdlOpType.SLL: " << ",
        HdlOpType.SRL: " >> ",

        HdlOpType.SLA: " <<< ",
        HdlOpType.SRA: " >>> ",

        HdlOpType.DOWNTO: ":",
        HdlOpType.TO: ":",
        HdlOpType.PART_SELECT_POST: " +: ",
        HdlOpType.PART_SELECT_PRE: " -: ",
        HdlOpType.MAP_ASSOCIATION: ":",

        HdlOpType.ARITH_SHIFT_LEFT_ASSIGN: ' <<<= ',
        HdlOpType.ARITH_SHIFT_RIGHT_ASSIGN: ' >>>= ',
        HdlOpType.DOUBLE_COLON: "::",
    }
    GENERIC_BIN_OPS.update(ToHdlCommon.GENERIC_BIN_OPS)
    GENERIC_BIN_OPS.update(ASSIGN_OPERATORS_SYMBOLS_C)

    OP_PRECEDENCE = {
        HdlOpType.APOSTROPHE: (1, L),
        HdlOpType.DOT: (1, L),
        HdlOpType.INDEX: (1, L),

        HdlOpType.CALL: (2, L),
        HdlOpType.TYPE_OF: (2, L),
        HdlOpType.PARAMETRIZATION: (2, L),

        HdlOpType.POW: (5, L),

        HdlOpType.CONCAT: (6, L),

        HdlOpType.REPL_CONCAT: (7, L),

        HdlOpType.DIV: (8, L),
        HdlOpType.MUL: (8, L),
        HdlOpType.MOD: (8, L),

        HdlOpType.ADD: (9, L),
        HdlOpType.SUB: (9, L),

        HdlOpType.SLL: (10, L),
        HdlOpType.SRL: (10, L),
        HdlOpType.SLA: (10, L),
        HdlOpType.SRA: (10, L),

        HdlOpType.GT: (11, L),
        HdlOpType.LT: (11, L),
        HdlOpType.GE: (11, L),
        HdlOpType.LE: (11, L),

        HdlOpType.EQ: (12, L),
        HdlOpType.NE: (12, L),
        HdlOpType.IS: (12, L),
        HdlOpType.IS_NOT: (12, L),
        HdlOpType.EQ_MATCH: (12, L),
        HdlOpType.NE_MATCH: (12, L),

        HdlOpType.AND: (13, L),
        HdlOpType.XOR: (13, L),
        HdlOpType.OR: (13, L),
        HdlOpType.NAND: (13, L),
        HdlOpType.XNOR: (13, L),

        HdlOpType.AND_LOG: (14, L),
        HdlOpType.OR_LOG: (14, L),

        HdlOpType.TERNARY: (15, R),

        HdlOpType.RISING: (16, R),
        HdlOpType.FALLING: (16, R),
        HdlOpType.DOWNTO: (17, L),
        HdlOpType.TO: (17, L),
        HdlOpType.PART_SELECT_PRE: (17, L),
        HdlOpType.PART_SELECT_POST: (17, L),
        HdlOpType.MAP_ASSOCIATION: (17, L),

    }
    OP_PRECEDENCE.update({k: (3, R) for k in [
        HdlOpType.MINUS_UNARY,
        HdlOpType.PLUS_UNARY,
        HdlOpType.INCR_PRE,
        HdlOpType.INCR_POST,
        HdlOpType.DECR_PRE,
        HdlOpType.DECR_POST,
        HdlOpType.NEG,
        HdlOpType.NEG_LOG,
        HdlOpType.OR_UNARY,
        HdlOpType.AND_UNARY,
        HdlOpType.NAND_UNARY,
        HdlOpType.NOR_UNARY,
        HdlOpType.XOR_UNARY,
        HdlOpType.XNOR_UNARY
    ]})
    OP_PRECEDENCE.update({k: (18, ASSOCIATIVITY.NONE)
                          for k in ASSIGN_OPERATORS})

    GENERIC_UNARY_OPS = {
        HdlOpType.NEG_LOG: "!",
        HdlOpType.NEG: "~",
        HdlOpType.MINUS_UNARY: "-",
        HdlOpType.PLUS_UNARY: "+",
        HdlOpType.OR_UNARY: "|",
        HdlOpType.AND_UNARY: "&",
        HdlOpType.NAND_UNARY: "~&",
        HdlOpType.NOR_UNARY: "~|",
        HdlOpType.XOR_UNARY: "^",
        HdlOpType.XNOR_UNARY: "~^",
        HdlOpType.RISING: "posedge ",
        HdlOpType.FALLING: "negedge ",
        HdlOpType.INCR_PRE: "++",
        HdlOpType.DECR_PRE: "--",
    }
    GENERIC_UNARY_OPS_POSTFIX = {
        HdlOpType.INCR_POST: "++",
        HdlOpType.DECR_POST: "--",
    }

    def visit_HdlValueInt(self, o):
        """
        :type o: HdlValueInt
        """
        w = self.out.write
        if o.bits is None:
            w(str(o.val))
        else:
            if o.base is None:
                f = "{0}'h{1:x}"
            else:
                b = o.base
                if b == 2:
                    base_char = 'b'
                elif b == 8:
                    base_char = 'O'
                elif b == 10:
                    base_char = 'd'
                elif b == 16:
                    base_char = 'h'
                else:
                    raise NotImplementedError(b)
                f = "{0}'" + base_char + "{1}"
            w(f.format(o.bits, o.val))

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        :return: True, the flag used to mark that the ; should be added if this is a statement
        """
        w = self.out.write
        if isinstance(o, HdlValueId):
            w(o.val)
        elif is_str(o):
            w('"%s"' % o)
        elif isinstance(o, HdlValueInt):
            self.visit_HdlValueInt(o)
        elif isinstance(o, HdlOp):
            _o, is_signed = pop_signed_flag(o)
            if o is not _o:
                self.visit_iHdlExpr(_o)
            else:
                self.visit_HdlOp(_o)

            if is_signed is True:
                w(" signed")
            elif is_signed is False:
                w(" unsigned")
        elif o is HdlAll:
            w("*")
        elif o is HdlTypeAuto:
            pass
        elif o is None:
            w("null")
        elif isinstance(o, float):
            w(str(o))
        elif isinstance(o, list):
            w("'{")
            for last, e in iter_with_last(o):
                self.visit_iHdlExpr(e)
                if not last:
                    w(", ")
            w("}")
        else:
            raise NotImplementedError(o.__class__, o)
        return True

    def _visit_operand(self, operand, i,
                           parent,
                           expr_requires_parenthesis,
                           cancel_parenthesis):

        if isinstance(operand, HdlOp) and \
               (operand.fn in self.GENERIC_UNARY_OPS or operand.fn in self.GENERIC_UNARY_OPS_POSTFIX) and \
               (parent.fn in self.GENERIC_UNARY_OPS or parent.fn in self.GENERIC_UNARY_OPS_POSTFIX):
            # unary operator in unary operator
            expr_requires_parenthesis = True

        return super(ToVerilog2005Expr, self)._visit_operand(operand, i,
                           parent,
                           expr_requires_parenthesis,
                           cancel_parenthesis)

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        """
        w = self.out.write
        op = o.fn
        if op == HdlOpType.CONCAT:
            w = self.out.write
            w("{")
            for is_last, (o_i, _o) in iter_with_last(enumerate(o.ops)):
                self._visit_operand(_o, o_i, o, False, True)
                if not is_last:
                    w(", ")
            w("}")
            return
        elif op == HdlOpType.REPL_CONCAT:
            w = self.out.write
            w("{")
            self._visit_operand(o.ops[0], 0, o, True, False)
            w("{")
            for is_last, (o_i, _o) in iter_with_last(enumerate(o.ops[1:])):
                self._visit_operand(_o, o_i, o, False, True)
                if not is_last:
                    w(", ")
            w("}}")
            return
        elif op == HdlOpType.TERNARY:
            cond, v0, v1 = o.ops
            self._visit_operand(cond, 0, o, True, False)
            w(" ? ")
            self._visit_operand(v0, 1, o, False, False)
            w(" : ")
            self._visit_operand(v1, 2, o, False, False)
            return
        elif op == HdlOpType.TYPE_OF:
            w("type(")
            self._visit_operand(o.ops[0], 0, o, False, True)
            w(")")
            return
        else:
            super(ToVerilog2005Expr, self).visit_HdlOp(o)
        return True

    def visit_type_first_part(self, t):
        """
        :type t: iHdlExpr
        :return: True if the type has also the array dimension part
        """
        w = self.out.write
        t, array_dims = collect_array_dims(t)
        wire_params = get_wire_t_params(t)
        if wire_params is None:
            if t != HdlTypeAuto:
                if isinstance(t, HdlOp) and t.fn == HdlOpType.TYPE_OF:
                    w("var ")
                self.visit_iHdlExpr(t)
            elif self._type_requires_nettype:
                w("wire")
        else:
            base_t, width, is_signed, _ = wire_params
            if base_t is not HdlTypeAuto:
                w(base_t.val)
            elif self._type_requires_nettype:
                w("wire")

            if is_signed is None:
                pass
            elif is_signed:
                w(" signed")
            else:
                w(" unsigned")

            if width is not None:
                # 1D vector
                w("[")
                self.visit_iHdlExpr(width)
                w("]")

        return len(array_dims) > 0

    def visit_type_array_part(self, t):
        """
        :type t: iHdlExpr
        """
        w = self.out.write
        _, array_dim = collect_array_dims(t)
        for ad in array_dim:
            w("[")
            if ad is not None:
                self.visit_iHdlExpr(ad)
            w("]")



from hdlConvertorAst.hdlAst import HdlIdDef, iHdlExpr, HdlOp, HdlOpType,\
    HdlDirection, HdlValueId, HdlStmProcess, HdlCompInst, HdlModuleDec,\
    HdlPhysicalDef, HdlEnumDef, HdlClassDef
from hdlConvertorAst.hdlAst._statements import ALL_STATEMENT_CLASSES
from hdlConvertorAst.to.basic_hdl_sim_model._main import ToBasicHdlSimModel
from hdlConvertorAst.to.common import ToHdlCommon
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last
from hdlConvertorAst.to.systemc.stm import ToSystemcStm
from hdlConvertorAst.hdlAst._expr import HdlTypeType
from hdlConvertorAst.to.verilog.utils import collect_array_dims
from hdlConvertorAst.py_ver_compatibility import method_as_function


DEFAULT_IMPORTS = """\
#include <systemc.h>
"""


class ToSystemc(ToSystemcStm):
    """
    Convert hdlObject AST to BasicHdlSimModel
    (Python simulation model used by hwtSimApi simulator)

    :ivar _is_param: flag which specifies if the current HdlIdDef is a param/generic
    :ivar _is_port: flag which specifies if the current HdlIdDef is a port
    """
    ALL_STATEMENT_CLASSES = ALL_STATEMENT_CLASSES

    def __init__(self, out_stream):
        ToHdlCommon.__init__(self, out_stream)
        self.module_path_prefix = None
        self.add_imports = True

    def visit_doc(self, obj):
        return super(ToSystemc, self).visit_doc(obj, "//")

    def visit_HdlModuleDef(self, mod_def):
        """
        :type mod_def: HdlModuleDef
        """
        mod_dec = mod_def.dec
        assert mod_dec is not None, mod_def
        assert not mod_dec.objs, mod_dec
        w = self.out.write
        w(DEFAULT_IMPORTS)
        w("\n")

        split_HdlModuleDefObjs = method_as_function(ToBasicHdlSimModel.split_HdlModuleDefObjs)
        types, variables, processes, components = \
            split_HdlModuleDefObjs(self, mod_def.objs)

        self.visit_doc(mod_dec)
        method_as_function(ToBasicHdlSimModel.visit_component_imports)(self, components)

        w("SC_MODULE(")
        w(mod_dec.name)
        w(") {\n")
        with Indent(self.out):
            if mod_dec.params:
                raise NotImplementedError()

            for t in types:
                self.visit_type_declr(t)
                w(";\n")

            w('// ports\n')
            try:
                self._is_port = True
                for p in mod_dec.ports:
                    self.visit_HdlIdDef(p)
                    w(";\n")
            finally:
                self._is_port = False

            w("// component instances\n")
            for c in components:
                w(c.module_name.val)
                w(" ")
                w(c.name.val)
                w('();\n')
            w("// internal signals\n")
            for v in variables:
                self.visit_HdlIdDef(v)
                w(";\n")

            for p in processes:
                self.visit_iHdlStatement(p)
                # extra line to separate a process functions
                w("\n")

            w("SC_CTOR(")
            w(mod_dec.name)
            w(")")
            if components:
                w(": ")
            for last, c in iter_with_last(components):
                w(c.name.val)
                w('("')
                w(c.name.val)
                if last:
                    w('")')
                else:
                    w('"), ')
            w(" {\n")
            with Indent(self.out):
                for p in processes:
                    if p.sensitivity:
                        w("SC_METHOD(")
                        w(p.labels[0])
                        w(");\n")
                        w("sensitive")
                        for s in p.sensitivity:
                            w(" << ")
                            self.visit_iHdlExpr(s)
                        w(";\n")
                    else:
                        w(p.labels[0])
                        w("();\n")
                w("// connect ports\n")
                for c in components:
                    for pm in c.port_map:
                        w(c.name.val)
                        w('.')
                        assert isinstance(pm, HdlOp) and\
                            pm.fn == HdlOpType.MAP_ASSOCIATION, pm
                        mod_port, connected_sig = pm.ops
                        assert isinstance(mod_port, HdlValueId), mod_port
                        self.visit_iHdlExpr(mod_port)
                        assert isinstance(
                            connected_sig, HdlValueId), connected_sig
                        w("(")
                        self.visit_iHdlExpr(connected_sig)
                        w(");\n")

            w("}\n")
        w("};\n")

    def visit_type_declr(self, var):
        """
        :type var: HdlIdDef
        """
        assert var.type == HdlTypeType
        self.visit_doc(var)
        w = self.out.write
        w("typedef ")
        t, arr_dims = collect_array_dims(var.value)
        with Indent(self.out):
            self.visit_iHdlExpr(t)
        w(" ")
        w(var.name)
        for d in arr_dims:
            w("[")
            self.visit_iHdlExpr(d)
            w("]")

        return True

    def visit_HdlIdDef(self, var):
        """
        :type var: HdlIdDef
        """
        self.visit_doc(var)
        w = self.out.write
        t, arr_dims = collect_array_dims(var.type)
        self.visit_type(t)
        w(" ")
        w(var.name)
        for d in arr_dims:
            w("[")
            self.visit_iHdlExpr(d)
            w("]")
        if var.value is not None:
            w(" = ")
            with Indent(self.out):
                self.visit_iHdlExpr(var.value)
        return True

    def visit_HdlClassDef(self, o):
        raise NotImplementedError()

    def visit_HdlEnumDef(self, o):
        raise NotImplementedError()

    def visit_HdlPhysicalDef(self, o):
        raise NotImplementedError()



from enum import Enum
from typing import Dict, Union, Tuple, List
from hdlConvertorAst.hdlAst._bases import iHdlObj,iHdlObjWithName

HdlTypeStr = str
# Verilog real
HdlTypeFloat = float


class iHdlTypeDef(iHdlObjWithName):
    """
    Base class for definitions of HDL types.

    :note: note that this is not a c-like typedef but the definition of a type itself.
        c-like typedef is described as shown in examples below.
        Equivalent of SystemVerilog/vhdl:

        .. code-block:: systemverilog

            typedef wire [7:0] octet;

        .. code-block:: vhdl

            subtype octet is std_logic_vector(7 downto 0); is

        .. code-block:: python

            HdlIdDef(name="octet", type=HdlTypeType, value=iHdlExpr(wire[7:0]))

    :note: vhdl arrays types are described as

        .. code-block:: python

            HdlIdDef("name", HdlTypeType,
               HdlOp(HdlOpType.INDEX, [type, array size])
    """


class HdlTypeBitsDef(iHdlTypeDef):
    """
    The type which represents bit or bit vector in HDL (std_logic/_vector
    in VHDL, [0:8] in Verilog )

    :ivar ~.msb: index of most significant bit
    :ivar ~.lsb: index of least significant bit
    :ivar ~.signed: True for SV/VDHL signed type, False for unsigned, None if signed/unsigned is not specified (VHDL std_logic_vector)
        None if not specified (std_logic_vector, wire, reg)
    :ivar ~.bit_order_bigendian: if True in VHDL std_logic_vector(msb downto lsb)
        becomes (msb to lsb)
    :ivar ~.force_vector: if msb==lsb==0 in VHDL this became std_logic
        instead of std_logic_vector(0 downto 0)
    :ivar ~.states: 2 means that each bit can be 0 or 1
        4 - (0, 1, X, Z) (e.g. Verilog wire)
        9 - (0, 1, X, Z, U, W, L, H, -) (e.g. VHDL std_logic)
            'U': uninitialized. This signal hasn't been set yet.
            'X': unknown. Impossible to determine this value/result.
            '0': logic 0
            '1': logic 1
            'Z': High Impedance
            'W': Weak signal, can't tell if it should be 0 or 1.
            'L': Weak signal that should probably go to 0
            'H': Weak signal that should probably go to 1
            '-': Don't care.

    :note: in new designs you should let lsb=0, it is there only for legacy issues
    """
    STD_LOGIC_STATES = 9
    WIRE_STATES = 4
    __slots__ = ["name", "msb", "lsb", "signed", "bit_order_bigendian", "states"]

    def __init__(self, msb, lsb=0, signed=False):
        super(HdlTypeBitsDef, self).__init__()
        self.msb = msb  # type: iHdlExpr
        self.lsb = lsb  # type: iHdlExpr
        self.signed = signed  # type: Optional[bool]
        self.bit_order_bigendian = False  # type: bool
        self.force_vector = True  # type: bool
        self.is_latched = False  # type: bool
        self.states = 2

    def width(self):
        if self.msb >= self.lsb:
            return self.msb - self.lsb
        else:
            return self.lsb - self.msb

    def __hash__(self):
        return hash((self.msb, self.lsb, self.signed, self.is_bigendian))

    def __eq__(self, other):
        return isinstance(other, HdlTypeBitsDef) and (
            self.msb == other.msb
            and self.lsb == other.lsb
            and self.signed == other.signed
            and self.is_bigendian == other.is_bigendian)


class HdlClassType(Enum):
    (
        CLASS,
        STRUCT,  # also vhdl record type
        UNION,
        INTERFACE,
    ) = range(4)


class HdlClassDef(iHdlTypeDef):
    """
    Definition of SystemVerilog class/struct/interface or VHDL record

    :ivar ~.base_types: parent classes for inheritance
    :type ~.base_types: List[iHdlExpr]
    :type type: HdlClassType
    :ivar ~.is_virtual: True for for SV abstract (virtual) classes
    :ivar ~.members: list of members
    :type ~.members: List[HdlIdDef]

    :note: equivalent of SystemVerilog/vhdl:

        .. code-block:: systemverilog

            typedef struct packed {
                int a;
            } s_name;

        .. code-block:: vhdl

            type s_name is record
                a: integer;
            end record;

        .. code-block:: python

            HdlIdDef(name="s_name", type=HdlTypeType,
                           value=HdlClassDef(...))

    """
    __slots__ = ["base_types", "is_virtual", "is_packed", "members"]

    def __init__(self):
        super(HdlClassDef, self).__init__()
        self.type = HdlClassType.CLASS
        self.base_types = []
        self.is_virtual = False  # type: bool
        self.is_packed = True  # type: bool
        self.members = []  # type: List[HdlIdDef]


class HdlPhysicalDef(iHdlTypeDef):
    """
    Definition of VHDL Physical Type

    :note: name may be None
    :note: equivalent of vhdl:

        .. code-block:: vhdl

           type p_name is range 0 to 100 
               units 
                   u1;
                   u2 = 10 u1;
                   u3 = 100 u2;
               end units;

        .. code-block:: python
           
           HdlIdDef(name="p_name", type=HdlTypeType,
                       value=HdlPhysicalDef(...))
    """
    __slots__ = ["name", "range", "members"]

    def __init__(self):
        super(HdlPhysicalDef, self).__init__()
        self.range = None
        self.members = []  # type: List[Tuple[str, Optional[iHdlExpr]]]


class HdlEnumDef(iHdlTypeDef):
    """
    Definition of VHDL Enumeration Type or SystemVerilog enum

    :note: name may be None
    :note: equivalent of SystemVerilog/vhdl:

        .. code-block:: systemverilog

            enum { a, b } e_name;

        .. code-block:: vhdl

           type e_name is (a, b);

        .. code-block:: python

            HdlIdDef(name="e_name", type=HdlTypeType,
                           value=HdlEnumDef("e_name", ["a", "b"]))
    """
    __slots__ = ["name", "values"]

    def __init__(self):
        super(HdlEnumDef, self).__init__()
        self.values = []  # type: List[Tuple[Optional[str], Optional[iHdlExpr]]]



from hdlConvertorAst.hdlAst import HdlIdDef, iHdlExpr, HdlOp, HdlOpType, \
    HdlDirection, HdlValueId, HdlStmProcess, HdlCompInst, HdlModuleDec, \
    HdlPhysicalDef, HdlEnumDef, HdlClassDef, HdlFunctionDef, iHdlTypeDef
from hdlConvertorAst.hdlAst._statements import ALL_STATEMENT_CLASSES
from hdlConvertorAst.py_ver_compatibility import method_as_function
from hdlConvertorAst.to.basic_hdl_sim_model._main import ToBasicHdlSimModel
from hdlConvertorAst.to.common import ToHdlCommon
from hdlConvertorAst.to.hdlUtils import Indent, iter_with_last
from hdlConvertorAst.to.hwt.stm import ToHwtStm

DEFAULT_IMPORTS = """\
from hwt.code import If, Switch, Concat
from hwt.code_utils import rename_signal
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import INT, SLICE, STR, BIT, FLOAT64
from hwt.hdl.types.enum import HEnum
from hwt.interfaces.std import Signal
from hwt.synthesizer.param import Param
from hwt.synthesizer.unit import Unit
"""


def pop_port_or_param_map(o):
    assert isinstance(o, HdlOp) and\
        o.fn == HdlOpType.MAP_ASSOCIATION, o
    mod_p, connected_p = o.ops
    assert isinstance(mod_p, HdlValueId), mod_p
    return mod_p, connected_p


def is_not_const(e, names_of_constants):
    """
    :param e: expression to check
    """
    if e is None:
        return False
    elif isinstance(e, HdlOp):
        for o in e.ops:
            if is_not_const(o, names_of_constants):
                return True
        return False
    elif isinstance(e, HdlValueId):
        return e.val not in names_of_constants
    return False


class ToHwt(ToHwtStm):
    """
    Convert hdlObject AST to BasicHdlSimModel
    (Python simulation model used by hwtSimApi simulator)

    :ivar _is_param: flag which specifies if the current HdlIdDef is a param/generic
    :ivar _is_port: flag which specifies if the current HdlIdDef is a port
    """
    ALL_STATEMENT_CLASSES = ALL_STATEMENT_CLASSES
    VAR_PER_LINE_LIMIT = 10
    TYPES_WHICH_CAN_BE_OMMITED = (HdlValueId("INT"), HdlValueId("STR"), HdlValueId("BOOL"), HdlValueId("FLOAT"))

    def __init__(self, out_stream):
        ToHdlCommon.__init__(self, out_stream)
        self.module_path_prefix = None
        self.add_imports = True
        self._is_port = False
        self._is_param = False

    def visit_doc(self, obj, doc_string=False):
        if doc_string:
            doc = obj.doc
            if doc is not None:
                doc = doc.split("\n")
                w = self.out.write
                if len(doc) > 1:
                    w('"""')
                    for d in doc:
                        w(d.replace('\r', ''))
                        w("\n")
                    w('"""\n')
                else:
                    w('"')
                    if doc:
                        w(doc[0].replace('\r', ''))
                    w('"\n')
        else:
            return super(ToHwt, self).visit_doc(obj, "#")

    def add_module_exampe_serialization(self, module_name):
        w = self.out.write
        w('if __name__ == "__main__":\n')
        with Indent(self.out):
            w("from hwt.synthesizer.utils import to_rtl_str\n")
            w("u = ")
            w(module_name)
            w("()\n")
            w("print(to_rtl_str(u))\n")

    def ivars_to_local_vars(self, var_names):
        if var_names:
            w = self.out.write
            VAR_PER_LINE_LIMIT = self.VAR_PER_LINE_LIMIT
            # ports and params to locals
            for i, (last, name) in enumerate(iter_with_last(var_names)):
                w(name)
                if not last:
                    w(", ")
                if i % VAR_PER_LINE_LIMIT == 0 and i > 0:
                    # jump to new line to have reasonably long variable list
                    w("\\\n")
            w(" = \\\n")
            for i, (last, name) in enumerate(iter_with_last(var_names)):
                w("self.")
                w(name)
                if not last:
                    w(", ")
                if i % VAR_PER_LINE_LIMIT == 0 and i > 0:
                        # jump to new line to have reasonably long variable list
                        w("\\\n")
            w("\n")

    def visit_HdlModuleDef(self, mod_def):
        """
        :type mod_def: HdlModuleDef
        """
        mod_dec = mod_def.dec
        assert mod_dec is not None, mod_def
        assert not mod_dec.objs, mod_dec
        w = self.out.write
        if self.add_imports:
            w(DEFAULT_IMPORTS)
            w("\n")
            if self.module_path_prefix is None:
                self.add_imports = False

        split_HdlModuleDefObjs = method_as_function(ToBasicHdlSimModel.split_HdlModuleDefObjs)
        otherDefs, variables, processes, components = \
            split_HdlModuleDefObjs(self, mod_def.objs)

        method_as_function(ToBasicHdlSimModel.visit_component_imports)(self, components)

        w("class ")
        w(mod_dec.name)
        w("(Unit):\n")
        with Indent(self.out):
            self.visit_doc(mod_dec, doc_string=True)

            params_names = []
            if mod_dec.params:
                w('def _config(self):\n')
                with Indent(self.out):
                    try:
                        self._is_param = True
                        for p in mod_dec.params:
                            self.visit_HdlIdDef(p, None)
                            params_names.append(p.name)
                    finally:
                        self._is_param = False
                w("\n")

            types = []
            for d in otherDefs:
                if isinstance(d, HdlFunctionDef):
                    self.visit_HdlFunctionDef(d)
                elif isinstance(d, iHdlTypeDef):
                    types.append(d)
                else:
                    raise NotImplementedError(d)

            names_of_constants = set(params_names)
            port_params_comp_names = [*params_names]
            w('def _declr(self):\n')
            with Indent(self.out):
                self.ivars_to_local_vars(port_params_comp_names)
                for t in types:
                    self.visit_type_declr(t)

                w('# ports\n')
                try:
                    self._is_port = True
                    for p in mod_dec.ports:
                        self.visit_HdlIdDef(p, names_of_constants)
                        port_params_comp_names.append(p.name)
                finally:
                    self._is_port = False

                w("# component instances\n")
                for c in components:
                    if c.param_map:
                        w(c.name.val)
                        w(" = ")
                    w("self.")
                    w(c.name.val)
                    w(" = ")
                    w(c.module_name.val)
                    w('()\n')
                    port_params_comp_names.append(c.name.val)
                    for cp in c.param_map:
                        mod_p, val = pop_port_or_param_map(cp)
                        w(c.name.val)
                        w(".")
                        w(mod_p.val)
                        w(" = ")
                        self.visit_iHdlExpr(val)
                        w("\n")

                w("\n")

            w("def _impl(self):\n")
            with Indent(self.out):
                self.ivars_to_local_vars(port_params_comp_names)
                w("# internal signals\n")
                for v in variables:
                    self.visit_HdlIdDef(v, names_of_constants)

                for c in components:
                    for pm in c.port_map:
                        mod_port, connected_sig = pop_port_or_param_map(pm)
                        assert isinstance(
                            connected_sig, HdlValueId), connected_sig
                        p = mod_port.obj
                        assert p is not None, (
                            "HdlValueId to module ports "
                            "shoudl have been discovered before")
                        d = p.direction
                        assert d.name in (HdlDirection.IN.name,
                                          HdlDirection.OUT.name), d
                        is_input = d.name == HdlDirection.IN.name
                        if is_input:
                            w(c.name.val)
                            w(".")
                            self.visit_iHdlExpr(mod_port)
                            w("(")
                            self.visit_iHdlExpr(connected_sig)
                            w(")\n")
                        else:
                            self.visit_iHdlExpr(connected_sig)
                            w("(")
                            w(c.name.val)
                            w(".")
                            self.visit_iHdlExpr(mod_port)
                            w(")\n")

                for p in processes:
                    self.visit_iHdlStatement(p)
                    # extra line to separate a process functions
                    w("\n")
        # w("\n")
        # w("\n")
        # self.add_module_exampe_serialization(mod_dec.name)

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        w = self.out.write
        w("def ")
        w(o.name)
        w("(self")
        for p in o.params:
            w(", ")
            w(p.name)
        w("):\n")
        with Indent(self.out):
            self.visit_doc(o, doc_string=True)
            for stm in o.body:
                self.visit_iHdlStatement(stm)
                w("\n")
        w("\n")

    def visit_HdlIdDef(self, var, names_of_constants):
        """
        :type var: HdlIdDef
        """
        self.visit_doc(var)
        w = self.out.write
        if self._is_port or self._is_param:
            w("self.")
        w(var.name)
        if self._is_port:
            w(" = Signal(")
            if var.type != HdlValueId("BIT"):
                self.visit_type(var.type)
            w(")")
            if var.direction == HdlDirection.OUT:
                w("._m()\n")
            else:
                w("\n")
                assert var.direction == HdlDirection.IN, var.direction
        elif self._is_param:
            w(" = Param(")
            if var.type in self.TYPES_WHICH_CAN_BE_OMMITED:
                # ommit the type specification in the case where the type can be resolved from value type
                self.visit_iHdlExpr(var.value)
                w(")\n")
            else:
                self.visit_type(var.type)
                w(".from_py(")
                self.visit_iHdlExpr(var.value)
                w("))\n")
        elif var.is_const:
            names_of_constants.add(var.name)
            w(" = ")
            if var.type in self.TYPES_WHICH_CAN_BE_OMMITED:
                self.visit_iHdlExpr(var.value)
                w("\n")
            else:
                self.visit_type(var.type)
                w(".from_py(")
                self.visit_iHdlExpr(var.value)
                w(")\n")
        else:
            # body signal
            if var.value is not None and var.value != HdlValueId("None") and is_not_const(var.value, names_of_constants):
                w(' = rename_signal(self, ')
                self.visit_iHdlExpr(var.value)
                w(', "')
                w(var.name)
                w('")\n')
            else:
                w(' = self._sig(')
                with Indent(self.out):
                    w('"')
                    w(var.name)
                    if var.type == HdlValueId("BIT"):
                        w('"')
                    else:
                        w('", ')
                        self.visit_type(var.type)
                if var.value is None:
                    w(")\n")
                else:
                    w(", def_val=")
                    self.visit_iHdlExpr(var.value)
                    w(")\n")




from hdlConvertorAst.hdlAst import HdlTypeBitsDef, HdlOpType, HdlOp, HdlValueId,\
    HdlTypeAuto, HdlValueInt
from hdlConvertorAst.to.basic_hdl_sim_model.utils import BitsT,\
    _verilog_slice_to_width
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_index
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.verilog_resolve_types import VerilogResolveTypes
from hdlConvertorAst.py_ver_compatibility import is_str


class VerilogTypesToBasicHdlSimModel(VerilogResolveTypes):
    """
    Translate Verilog HDL types to BasicHdlSimModel HDL types
    """
    def _visit_type(self, t):
        """
        :type t: Union[iHdlExpr, iHdlTypeDef]
        """
        if isinstance(t, HdlOp) and t.fn == HdlOpType.INDEX:
            o0, o1 = t.ops
            return hdl_index(self._visit_type(o0), o1)
        elif isinstance(t, HdlTypeBitsDef):
            width = _verilog_slice_to_width(t.msb, t.lsb)
            return BitsT(width, bool(t.signed))
        elif isinstance(t, HdlValueId):
            if t == HdlValueId("integer"):
                return HdlValueId("INT")
        raise NotImplementedError(t)

    def visit_HdlIdDef(self, o):
        """
        :type o: HdlIdDef
        """
        if o.type is HdlTypeAuto:
            if is_str(o.value):
                o.type = HdlValueId("STR")
                return o
            elif isinstance(o.value, HdlValueInt):
                v = o.value.val
                if isinstance(v, int) and (v <= 1 and v >= 0):
                    o.type =  HdlValueId("BIT")
                else:
                    o.type =  HdlValueId("INT")
                return o

        return VerilogResolveTypes.visit_HdlIdDef(self, o)

    def visit_type(self, t):
        """
        :type t: iHdlExpr
        """
        t = VerilogResolveTypes.visit_type(self, t)
        return self._visit_type(t)




class CodePosition(object):
    __slots__ = [
        "start_line",
        "start_column",
        "stop_line",
        "stop_column",
    ]

    def __init__(self):
        self.start_line = None
        self.start_column = None
        self.stop_line = None
        self.stop_column = None

    def __repr__(self):
        return "<%s %r:%r to %r:%r>" % (self.__class__.__name__,
                                        self.start_line, self.start_column,
                                        self.stop_line, self.stop_column)



import sys

# [TODO] merge with py_ver_independent_str

if sys.version_info[0] <= 2:
    def is_str(x):
        return isinstance(x, (basestring, unicode))
else:
    def is_str(x):
        return isinstance(x, str)


def method_as_function(fn):
    if sys.version_info[0] <= 2:
        return fn.im_func
    return fn



from hdlConvertorAst.hdlAst import HdlValueId, HdlOp, HdlOpType, HdlTypeAuto


PRIMITIVE_TYPES = (
    HdlValueId("reg"),
    HdlValueId("wire"),
    HdlValueId("bit"),
    HdlValueId("logic"),
    HdlValueId("signed"),
    HdlValueId("unsigned"),
    HdlTypeAuto,
)

NULL = HdlValueId("null")


def collect_array_dims(t):
    array_dim = []
    while isinstance(t, HdlOp) and t.fn == HdlOpType.INDEX:
        assert len(t.ops) <= 2
        if len(t.ops) == 2:
            d = t.ops[1]
        else:
            d = None
        array_dim.append(d)
        t = t.ops[0]
    array_dim.reverse()
    return t, array_dim


def get_wire_t_params(t):
    """
    wire/reg type is actually stored as: t#(width, is_signed)
    This function extracts t, width, is_signed and potential array dimmensions
    if this type is an array.
    """
    t, array_dim = collect_array_dims(t)

    if t in PRIMITIVE_TYPES:
        is_signed = None
        if t == HdlValueId("signed"):
            is_signed = True
        elif t == HdlValueId("unsigned"):
            is_signed = False
        return t, None, is_signed, array_dim

    # 1b scala
    if not isinstance(t, HdlOp) or t.fn != HdlOpType.PARAMETRIZATION or len(t.ops) != 3:
        return None

    if t.ops[0] not in PRIMITIVE_TYPES:
        return None

    t, width, is_signed = t.ops
    if width == NULL:
        width = None
    if is_signed is None:
        is_signed = None
    else:
        is_signed = bool(is_signed)

    return t, width, is_signed, array_dim



from hdlConvertorAst.to.verilog.keywords import IEEE1800_2017_DOLAR_SYMBOLS
from hdlConvertorAst.translate.common.discover_declarations import BuiltIn


def propopulate_verilog_builtins(name_scope):
    """
    :type name_scope: NameScope
    """
    for kw in IEEE1800_2017_DOLAR_SYMBOLS:
        name_scope.register_name(kw, BuiltIn(kw))

    name_scope.register_name('void', BuiltIn('void'))



from enum import Enum
from typing import List, Optional, Union, Tuple

from hdlConvertorAst.hdlAst._bases import iHdlStatement, iHdlObj
from hdlConvertorAst.hdlAst._expr import iHdlExpr


class HdlImport(iHdlStatement):
    """
    The import statements used in VHDL

    :note: note that this does not corresponds to include as include
            is processed by preprocessor
    :ivar ~.path: the list of names which are in the import
    """
    __slots__ = ["path", ]

    def __init__(self, path):
        """
        :type path: List[Union[HdlValueId, HdlAll]]
        """
        super(HdlImport, self).__init__()
        self.path = path


class HdlStmNop(iHdlStatement):
    """
    Nop statement in HDL
    """
    __slots__ = []


class HdlStmBlockJoinType(Enum):
    """
    Python equivalent of C++ hdlConvertor::hdlAst::HdlStmBlockJoinType
    """
    (
        SEQ,  # sequential block
        PAR_JOIN,  # parallel block where we are wainting on all threads
        PAR_JOIN_ANY,  # parallel block where we are wainting on first finished thread
        PAR_JOIN_NONE,  # parallel block where we are are not waiting on any thread
    ) = range(4)


class HdlStmBlock(iHdlStatement):
    """
    Block of statements in HDL
    """
    __slots__ = ["join_t", "body", ]

    def __init__(self):
        super(HdlStmBlock, self).__init__()
        self.join_t = HdlStmBlockJoinType.SEQ  # type: HdlStmBlockJoinType
        self.body = []  # type: List[iHdlObj]


class HdlStmAssign(iHdlStatement):
    """
    HDL assignment statement

    :ivar ~.src: the source
    :ivar ~.dst: the destination
    :ivar ~.is_blocking: flag if true assignment should be evaluated immediately
        else the assignment is evaluated after end of the current event

    :ivar ~.time_delay: delay which waits for a specified time
    :ivar ~.event_delay: delay which waits for a specified event
    :note: Evaluation of assignments with delay specified
           is generaly not synthetisable.
           If statement is blocking the current statement waits
           until the condition is met and then continues.
           If the assignment is not blocking the code continues
           and this statement is executed asynchronously
           after condition is met.
    """
    __slots__ = ["src", "dst", "is_blocking", "time_delay", "event_delay"]

    def __init__(self, src, dst, time_delay=None, event_delay=None):
        super(HdlStmAssign, self).__init__()
        self.src = src  # type: iHdlExpr
        self.dst = dst  # type: iHdlExpr
        self.is_blocking = False  # type: bool
        self.time_delay = time_delay  # type Optional[iHdlExpr]
        self.event_delay = event_delay  # type Optional[List[iHdlExpr]]


class HdlStmIf(iHdlStatement):
    """
    HDL if statement

    :ivar ~.cond: condition in if statement
    :ivar ~.if_true: block of statements which is in if true branch
    :ivar ~.elifs: type: List[Tuple[iHdlExpr, iHdlStatement]]
                  = list of (condition, statement list)
    """
    __slots__ = ["cond", "if_true", "elifs", "if_false"]

    def __init__(self):
        super(HdlStmIf, self).__init__()
        self.cond = None  # type: iHdlExpr
        self.if_true = None  # type: iHdlStatement
        self.elifs = []  # type: List[Tuple[iHdlExpr, iHdlStatement]]
        self.if_false = None  # type: Optional[iHdlStatement]

class HdlStmProcessTriggerConstrain(Enum):
    """
    This enum is used to mark SystemVerilog always construct type variants.

    :note: just always is marked with None, this enum specifies only special always construct variants
        namely always_comb, always_ff, always_latch
    """
    (COMB, FF, LATCH) = range(3)


class HdlStmProcess(iHdlStatement):
    """
    HDL process statement
    the container of statements with the sensitivity specified

    :ivar ~.sensitivity: optional list of expressions which specifies
        the trigger signals of the evaluation of the process
    :note: sensitivity = None means that the process is executed immediately
        sensitivity = [] means the process is never executed
    :note: Verilog always #time construct is translated to process
           without sensitivity and wait #time as a first statement in body.
    """
    __slots__ = ["sensitivity", "body", "trigger_constrain"]

    def __init__(self):
        super(HdlStmProcess, self).__init__()
        self.sensitivity = None  # type: Optional[List[iHdlExpr]]
        self.trigger_constrain = None # type: Optional[HdlStmProcessTriggerConstrain]
        self.body = None  # type: iHdlStatement


class HdlStmCaseType(Enum):
    (CASE, CASEX, CASEZ) = range(3)

class HdlStmCaseUniqConstrain(Enum):
    """
    Used to desinguis between system verilog case variants
    """
    (UNIQUE, UNIQUE0, PRIORITY) = range(3)

class HdlStmCase(iHdlStatement):
    """
    HDL case statement
    """
    __slots__ = ["type", "switch_on", "cases", "default", "uniq_constrain"]

    def __init__(self):
        super(HdlStmCase, self).__init__()
        self.type = HdlStmCaseType.CASE
        self.switch_on = None  # type: iHdlExpr
        self.uniq_constrain = None # type: Optional[HdlStmCaseUniqConstrain]
        self.cases = []  # type: List[Tuple[iHdlExpr, iHdlStatement]]
        self.default = None  # type: Optional[iHdlStatement]


class HdlStmFor(iHdlStatement):
    """
    HDL for statement

    for (init, cond, step)
        body
    """
    __slots__ = ["init", "cond", "step", "body"]

    def __init__(self):
        super(HdlStmFor, self).__init__()
        self.init = None  # type: iHdlStatement
        self.cond = None  # type: iHdlExpr
        self.step = None  # type: iHdlStatement
        self.body = None  # type: iHdlStatement


class HdlStmForIn(iHdlStatement):
    """
    HDL for in statement

    for var in collection:
        body

    :note: corresponds to VHDL "for" and is used also for Verilog "repeat"
    """
    __slots__ = ["var_defs", "collection", "body"]

    def __init__(self):
        super(HdlStmForIn, self).__init__()
        self.var_defs = []  # type: List[iHdlObj]
        self.collection = None  # type: iHdlExpr
        self.body = None  # type: iHdlStatement


class HdlStmWhile(iHdlStatement):
    """
    HDL while statement

    while cond:
        body
    """
    __slots__ = ["cond", "body"]

    def __init__(self):
        super(HdlStmWhile, self).__init__()
        self.cond = None  # type: iHdlExpr
        self.body = None  # type: iHdlStatement


class HdlStmRepeat(iHdlStatement):
    """
    HDL repeat statement

    for _ in range(n):
        body
    """
    __slots__ = ["n", "body"]

    def __init__(self):
        super(HdlStmRepeat, self).__init__()
        self.n = None  # type: iHdlExpr
        self.body = None  # type: iHdlStatement


class HdlStmReturn(iHdlStatement):
    """
    HDL return statement
    """
    __slots__ = ["val"]

    def __init__(self):
        super(HdlStmReturn, self).__init__()
        self.val = None  # type: iHdlExpr


class HdlStmWait(iHdlStatement):
    """
    HDL wait statement

    This is generally not synthetisable to hardware and is used in simulations.
    """
    __slots__ = ["val"]

    def __init__(self):
        super(HdlStmWait, self).__init__()
        self.val = []  # type: List[iHdlExpr]


class HdlStmBreak(iHdlStatement):
    """
    HDL break statement
    """
    __slots__ = []


class HdlStmContinue(iHdlStatement):
    """
    HDL continue statement
    """
    __slots__ = []


class HdlStmThrow(iHdlStatement):
    """
    HDL throw (raise) statement
    """
    __slots__ = ["val"]

    def __init__(self):
        super(HdlStmThrow, self).__init__()
        self.val = None  # type: iHdlExpr


ALL_STATEMENT_CLASSES = (
    HdlStmNop,
    HdlStmBlock,
    HdlStmAssign,
    HdlStmIf,
    HdlStmProcess,
    HdlStmCase,
    HdlStmFor,
    HdlStmForIn,
    HdlStmWhile,
    HdlStmRepeat,
    HdlStmReturn,
    HdlStmWait,
    HdlStmBreak,
    HdlStmContinue,
    HdlStmThrow,
)



from hdlConvertorAst.hdlAst import HdlDirection, HdlValueId, HdlValueInt, \
    HdlOp, NON_INSTANCIABLE_NODES
from hdlConvertorAst.py_ver_compatibility import is_str
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.hdlAst._typeDefs import iHdlTypeDef


class ToJson(HdlAstVisitor):

    def visit_HdlContext(self, context):
        """
        :type context: HdlContext
        """
        res = []
        for o in context.objs:
            d = self.visit_main_obj(o)
            res.append(d)
        return res

    def visit_HdlLibrary(self, o):
        """
        :type o: HdlLibrary
        """
        return self.visit_iHdlObjWithName(o)

    def visit_iHdlObjWithName(self, o):
        """
        :type o: iHdlObj
        """
        d = self.visit_iHdlObj(o)
        d["name"] = self.visit_iHdlExpr(o.name)
        return d

    def visit_CodePosition(self, o):
        """
        :type o: CodePosition
        """
        return (o.start_line, o.start_column, o.stop_line, o.stop_column)

    def visit_iHdlObj(self, o):
        """
        :type o: iHdlObj
        """
        d = {
            "__class__": o.__class__.__name__,
        }
        if o.doc:
            d["doc"] = self.visit_iHdlExpr(o.doc)
        if o.position:
            d["position"] = self.visit_CodePosition(o.position)
        return d

    def visit_HdlImport(self, o):
        """
        :type o: HdlImport
        """
        d = self.visit_iHdlObj(o)
        d["path"] = self.visit_iHdlExpr(o.path)
        return d

    def visit_HdlValueIdspace(self, o):
        """
        :type o: HdlValueIdspace
        """
        d = self.visit_iHdlObjWithName(o)
        d["objs"] = [self.visit_main_obj(o2) for o2 in o.objs]
        d["declaration_only"] = o.declaration_only
        return d

    def visit_HdlDirection(self, o):
        """
        :type o: HdlDirection
        """
        if o is None or o == HdlDirection.UNKNOWN:
            return None
        else:
            return o.name

    def visit_HdlIdDef(self, o):
        """
        :type o: HdlIdDef
        """
        d = self.visit_iHdlObjWithName(o)

        for flag_name in ["is_latched", "is_const", "is_static",
                          "is_virtual", ]:
            if getattr(o, flag_name):
                d[flag_name] = True
        d["type"] = self.visit_iHdlExpr(o.type)
        if o.value is not None:
            d["value"] = self.visit_iHdlExpr(o.value)
        d["direction"] = self.visit_HdlDirection(o.direction)
        return d

    def visit_HdlModuleDec(self, o):
        """
        :type o: HdlModuleDec
        """
        d = self.visit_HdlValueIdspace(o)
        d["params"] = [self.visit_HdlIdDef(p) for p in o.params]
        d["ports"] = [self.visit_HdlIdDef(p) for p in o.ports]
        return d

    def visit_HdlModuleDef(self, o):
        """
        :type o: HdlModuleDec
        """
        d = self.visit_iHdlObjWithName(o)
        if o.dec is not None:
            d["dec"] = self.visit_HdlModuleDec(o.dec)
        if o.module_name is not None:
            d["module_name"] = self.visit_iHdlExpr(o.module_name)
        d["objs"] = [self.visit_main_obj(o2) for o2 in o.objs]
        return d

    def _visit_iHdlStatement(self, o):
        """
        :type o: iHdlStatement
        """
        d = self.visit_iHdlObj(o)
        d["labels"] = [self.visit_iHdlExpr(str(x)) for x in o.labels]
        if o.in_preproc:
            d["in_preproc"] = True
        return d

    def visit_HdlCompInst(self, o):
        """
        :type o: HdlCompInst
        """
        d = self.visit_iHdlObjWithName(o)
        if o.module_name is not None:
            d["module_name"] = self.visit_iHdlExpr(o.module_name)
        d["param_map"] = [self.visit_iHdlExpr(pm) for pm in o.param_map]
        d["port_map"] = [self.visit_iHdlExpr(pm) for pm in o.port_map]
        return d

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        d = self.visit_iHdlObjWithName(o)
        for f in ["is_declaration_only",
                  "is_operator",
                  "is_static",
                  "is_task",
                  "is_virtual",
                  ]:
            if getattr(o, f):
                d[f] = True
        if o.return_t is not None:
            d["return_t"] = self.visit_iHdlExpr(o.return_t)
        d["params"] = [self.visit_HdlIdDef(v) for v in o.params]
        d["body"] = [self.visit_main_obj(o2) for o2 in o.body]
        return d

    def visit_HdlClassDef(self, o):
        """
        :type o: HdlClassDef
        """
        d = self.visit_iHdlObjWithName(o)
        d["members"] = [HdlAstVisitor.visit_iHdlObj(self, m) for m in o.members]
        d["type"] = o.type.name
        d["base_types"] = [self.visit_iHdlExpr(t) for t in o.base_types]
        d["is_virtual"] = o.is_virtual
        d["is_packed"] = o.is_packed
        return d

    def visit_HdlPhysicalDef(self, o):
        """
        :type o: HdlPhysicalDef
        """
        d = self.visit_iHdlObjWithName(o)
        d["range"] = self.visit_iHdlExpr(o.range)
        d["members"] = [self.visit_iHdlExpr(v) for v in o.members]
        return d

    def visit_HdlEnumDef(self, o):
        """
        :type o: HdlEnumDef
        """
        d = self.visit_iHdlObjWithName(o)
        d["values"] = [self.visit_iHdlExpr(v) for v in o.values]
        return d

    def visit_HdlStmProcess(self, o):
        """
        :type o: HdlStmProcess
        """
        d = self._visit_iHdlStatement(o)
        if o.sensitivity is not None:
            d["sensitivity"] = [self.visit_iHdlExpr(x) for x in o.sensitivity]
        if o.trigger_constrain is not None:
            d["trigger_constrain"] = o.trigger_constrain.name
        d["body"] = self.visit_iHdlStatement(o.body)
        return d

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        d = self._visit_iHdlStatement(o)
        d["cond"] = self.visit_iHdlExpr(o.cond)
        d["if_true"] = self.visit_iHdlStatement(o.if_true)
        d["elifs"] = [
            [self.visit_iHdlExpr(c), self.visit_iHdlStatement(stm)]
            for c, stm in o.elifs
        ]
        if o.if_false is not None:
            d["if_false"] = self.visit_iHdlStatement(o.if_false)
        return d

    def visit_HdlStmBlockJoinType(self, o):
        """
        :type o: HdlStmBlockJoinType
        """
        return o.name

    def visit_HdlStmBlock(self, o):
        """
        :type o: HdlStmBlock
        """
        d = self._visit_iHdlStatement(o)
        d["join_t"] = self.visit_HdlStmBlockJoinType(o.join_t)
        d["body"] = [self.visit_main_obj(o2) for o2 in o.body]
        return d

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase

        :return: True if requires ;\\n after end
        """
        d = self._visit_iHdlStatement(o)
        d["type"] = o.type.name
        d["switch_on"] = self.visit_iHdlExpr(o.switch_on)
        d["cases"] = [
            [self.visit_iHdlExpr(c), self.visit_iHdlStatement(stm)]
            for c, stm in o.cases
        ]
        if o.default is not None:
            d["default"] = self.visit_iHdlStatement(o.default)
        return d

    def visit_HdlStmWait(self, o):
        """
        :type o: HdlStmWait
        """
        d = self._visit_iHdlStatement(o)
        d["val"] = self.visit_iHdlExpr(o.val)
        return d

    def visit_HdlStmFor(self, o):
        """
        :type o: HdlStmFor
        """
        d = self._visit_iHdlStatement(o)
        d["init"] = self.visit_iHdlStatement(o.init)
        d["cond"] = self.visit_iHdlExpr(o.cond)
        d["step"] = self.visit_iHdlStatement(o.step)
        d["body"] = self.visit_iHdlStatement(o.body)
        return d

    def visit_HdlStmForIn(self, o):
        """
        :type o: HdlStmForIn
        """
        d = self._visit_iHdlStatement(o)
        d["var_defs"] = [self.visit_main_obj(o2) for o2 in o.var_defs]
        d["collection"] = self.visit_iHdlExpr(o.collection)
        d["body"] = self.visit_iHdlStatement(o.body)
        return d

    def visit_HdlStmWhile(self, o):
        """
        :type o: HdlStmWhile
        """
        d = self._visit_iHdlStatement(o)
        d["cond"] = self.visit_iHdlExpr(o.cond)
        d["body"] = self.visit_iHdlStatement(o.body)
        return d

    def visit_HdlStmRepeat(self, o):
        """
        :type o: HdlStmRepeat
        """
        d = self._visit_iHdlStatement(o)
        d["n"] = self.visit_iHdlExpr(o.n)
        d["body"] = self.visit_iHdlStatement(o.body)
        return d

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        d = self._visit_iHdlStatement(o)
        if o.event_delay is not None:
            d["event_delay"] = self.visit_iHdlExpr(o.event_delay)
        if o.time_delay is not None:
            d["time_delay"] = self.visit_iHdlExpr(o.time_delay)
        d["src"] = self.visit_iHdlExpr(o.src)
        d["dst"] = self.visit_iHdlExpr(o.dst)
        d["is_blocking"] = bool(o.is_blocking)
        return d

    def visit_HdlStmReturn(self, o):
        """
        :type o: HdlStmReturn
        """
        d = self._visit_iHdlStatement(o)
        if o.val is not None:
            d["val"] = self.visit_iHdlExpr(o.val)
        return d

    def visit_HdlStmContinue(self, o):
        """
        :type o: HdlStmContinue
        """
        return self._visit_iHdlStatement(o)

    def visit_HdlStmBreak(self, o):
        """
        :type o: HdlStmBreak
        """
        return self._visit_iHdlStatement(o)

    def visit_HdlStmThrow(self, o):
        """
        :type o: HdlStmThrow
        """
        d = self._visit_iHdlStatement(o)
        if o.val is not None:
            d["val"] = self.visit_iHdlExpr(o.val)
        return d

    def visit_HdlStmNop(self, o):
        """
        :type o: HdlStmNop
        """
        d = self._visit_iHdlStatement(o)
        return d

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        :return: iHdlExpr
        """
        if isinstance(o, HdlValueId):
            return o.val
        elif is_str(o) or o is None:
            d = {
                "__class__": "str",
                "val": o
            }
        elif isinstance(o, HdlValueInt):
            d = self.visit_HdlValueInt(o)
        elif isinstance(o, HdlOp):
            d = self.visit_HdlOp(o)
        elif isinstance(o, (list, tuple, dict)):
            if isinstance(o, dict):
                items = []
                for _k, _v in o.items():
                    k = self.visit_iHdlExpr(_k)
                    v = self.visit_iHdlExpr(_v)
                    items.append((k, v))
            else:
                items = [self.visit_iHdlExpr(o2) for o2 in o]

            d = {
                "__class__": o.__class__.__name__,
                "items": items,
            }
        elif isinstance(o, iHdlTypeDef):
            return HdlAstVisitor.visit_iHdlObj(self, o)
        elif isinstance(o, float):
            return o
        elif o in NON_INSTANCIABLE_NODES:
            d = {
                "__class__": o.__name__,
            }
        else:
            raise NotImplementedError(
                "Unexpected object of type " + str(type(o)))
        return d

    def visit_HdlValueInt(self, o):
        """
        :type o: HdlValueInt
        """
        if isinstance(o.val, int) and o.bits is None and o.base is None:
            return o.val

        d = {
            "__class__": o.__class__.__name__,
            "val": o.val,
        }
        if o.bits is not None:
            d["bits"] = o.bits
        if o.base is not None:
            d["base"] = o.base
        return d

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        :return: iHdlExpr
        """
        d = {
            "__class__": o.__class__.__name__,
            "ops": [self.visit_iHdlExpr(op) for op in o.ops],
            "fn": o.fn.name
        }
        return d



from hdlConvertorAst.hdlAst._structural import HdlContext, HdlModuleDef,\
    HdlModuleDec
from hdlConvertorAst.hdlAst._expr import HdlValueId


def link_module_dec_def(context):
    """
    :type context: HdlContext
    """
    objs = []
    last = None
    for o in context.objs:
        if isinstance(o, HdlModuleDef) and o.dec is None:
            assert isinstance(last, HdlModuleDec) \
                and o.module_name == HdlValueId(last.name), (
                "Module body has to be behind the module header", last, o)
            o.dec = last
            objs.append(o)
            last = None
        else:
            if last is not None:
                objs.append(last)
            last = o
    if last is not None:
        objs.append(last)

    context.objs = objs



from hdlConvertorAst.translate.common.add_call_operator_for_call_without_parenthesis import AddCallOperatorForCallWithoutParenthesis
from hdlConvertorAst.translate.common.discover_declarations import DiscoverDeclarations
from hdlConvertorAst.translate.common.name_scope import NameScope
from hdlConvertorAst.translate.common.resolve_names import ResolveNames
from hdlConvertorAst.translate.verilog_builtins import propopulate_verilog_builtins
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.detect_compiletime_statements import DetectCompileTimeStatements
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model\
    .verilog_operands_to_basic_hdl_sim_model import BasicHdlSimModelTranslateVerilogOperands
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model\
   .wrap_module_statements_to_processes import wrap_module_statements_to_processes
from hdlConvertorAst.translate.verilog_to_hwt.signal_assignments_to_call_op import SignalAssignmentsToCallOp
from hdlConvertorAst.translate.verilog_to_hwt.verilog_types_to_hwt import VerilogTypesToHwt
from hdlConvertorAst.translate.verilog_to_vhdl.inject_process_sens_to_statements import InjectProcessSensToStatements
from hdlConvertorAst.translate.vhdl_to_verilog import link_module_dec_def


def verilog_to_hwt(context):
    """
    :type context: HdlContext
    """
    link_module_dec_def(context)
    name_scope = NameScope.make_top(False)
    propopulate_verilog_builtins(name_scope)

    DiscoverDeclarations(name_scope).visit_HdlContext(context)
    ResolveNames(name_scope).visit_HdlContext(context)
    DetectCompileTimeStatements().visit_HdlContext(context)
    InjectProcessSensToStatements().visit_HdlContext(context)
    BasicHdlSimModelTranslateVerilogOperands(downto_to_slice_fn=False).visit_HdlContext(context)
    VerilogTypesToHwt().visit_HdlContext(context)
    AddCallOperatorForCallWithoutParenthesis().visit_HdlContext(context)
    wrap_module_statements_to_processes(context)
    SignalAssignmentsToCallOp().visit_HdlContext(context)

    return context, name_scope



from hdlConvertorAst.hdlAst import HdlImport, HdlStmProcess, HdlStmIf,\
    HdlStmAssign, HdlStmCase, HdlStmWait, HdlStmReturn, HdlStmFor, HdlStmForIn,\
    HdlStmWhile, HdlStmBlock, iHdlStatement, HdlModuleDec, HdlModuleDef,\
    HdlValueIdspace, HdlIdDef, HdlFunctionDef, HdlOp, HdlCompInst, \
    HdlValueInt, HdlStmBreak, HdlStmContinue, HdlStmRepeat, HdlLibrary, HdlContext,\
    HdlClassDef, HdlPhysicalDef, HdlEnumDef
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor


class HdlAstModifier(HdlAstVisitor):
    """
    A visitor which can be used to traverse and modyfy AST (Abstract Syntax Tree) made of objects
    from `hdlConvertorAst.hdlAst` module.
    Each visit function has to return the object which replaces current object
    It should return the same object if no change is required.
    """

    def visit_iHdlObj_list(self, obj_list, visit_fn):
        """
        :ivar obj_list: list of HDL objects
        :ivar visit_fn: function to modidify the object
        """
        for i, o in enumerate(obj_list):
            obj_list[i] = visit_fn(o)
        return obj_list

    def visit_HdlContext(self, context):
        """
        :type context: HdlContext
        """
        self.visit_iHdlObj_list(context.objs, self.visit_main_obj)
        return context

    def visit_HdlValueIdspace(self, o):
        """
        :type o: HdlValueIdspace
        """
        self.visit_doc(o)
        self.visit_iHdlObj_list(o.objs, self.visit_iHdlObj)
        return o

    def visit_HdlModuleDec(self, o):
        """
        :type o: HdlModuleDec
        """
        self.visit_doc(o)
        self.visit_iHdlObj_list(o.params, self.visit_param)
        self.visit_iHdlObj_list(o.ports, self.visit_port)
        self.visit_iHdlObj_list(o.objs, self.visit_main_obj)
        return o

    def visit_HdlIdDef(self, o):
        """
        :type o: HdlIdDef
        """
        self.visit_doc(o)
        o.type = self.visit_type(o.type)
        if o.value is not None:
            o.value = self.visit_iHdlExpr(o.value)
        
        return o

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        :return: iHdlExpr
        """
        if isinstance(o, HdlOp):
            return self.visit_HdlOp(o)
        else:
            return o

    def visit_HdlOp(self, o):
        """
        :type o: HdlOp
        :return: iHdlExpr
        """
        self.visit_iHdlObj_list(o.ops, self.visit_iHdlExpr)
        return o

    def visit_HdlValueInt(self, o):
        """
        :type o: HdlValueInt
        """
        return o

    def visit_port(self, o):
        return self.visit_HdlIdDef(o)

    def visit_param(self, o):
        return self.visit_HdlIdDef(o)

    def visit_HdlModuleDef(self, o):
        """
        :type o: HdlModuleDef
        """
        if o.dec is not None:
            o.dec = self.visit_HdlModuleDec(o.dec)

        self.visit_doc(o)
        objs = o.objs
        for i, _o in enumerate(objs):
            if isinstance(_o, iHdlStatement):
                _o = self.visit_iHdlStatement(_o)
            elif isinstance(_o, HdlIdDef):
                _o = self.visit_HdlIdDef(_o)
            elif isinstance(_o, HdlCompInst):
                _o = self.visit_HdlCompInst(_o)
            elif isinstance(_o, HdlFunctionDef):
                _o = self.visit_HdlFunctionDef(_o)
            else:
                raise NotImplementedError(_o)
            objs[i] = _o
        return o

    def visit_HdlCompInst(self, o):
        """
        :type o: HdlCompInst
        """
        self.visit_doc(o)
        self.visit_iHdlObj_list(o.param_map, self.visit_iHdlExpr)
        self.visit_iHdlObj_list(o.port_map, self.visit_iHdlExpr)
        return o

    def visit_HdlClassDef(self, o):
        """
        :type o: HdlClassDef
        """
        self.visit_doc(o)
        
        self.visit_iHdlObj_list(o.base_types, self.visit_iHdlExpr)
        self.visit_iHdlObj_list(o.members, self.visit_iHdlObj)

        return o

    def visit_HdlPhysicalDef(self, o):
        """
        :type o: HdlPhysicalDef
        """
        return o
    
    def visit_HdlEnumDef(self, o):
        """
        :type o: HdlEnumDef
        """
        return o

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        self.visit_doc(o)
        self.visit_iHdlObj_list(o.params, self.visit_HdlIdDef)
        if o.return_t is not None:
            o.return_t = self.visit_iHdlExpr(o.return_t)

        self.visit_iHdlObj_list(o.body, self.visit_main_obj)
        return o

    def visit_HdlStmProcess(self, o):
        """
        :type o: HdlStmProcess
        """
        self.visit_doc(o)
        o.body = self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmBlock(self, o):
        """
        :type o: HdlStmBlock
        """
        self.visit_doc(o)
        self.visit_iHdlObj_list(o.body, self.visit_iHdlStatement)
        return o

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase
        """
        self.visit_doc(o)
        o.switch_on = self.visit_iHdlExpr(o.switch_on)
        cases = o.cases
        for i, (c, stm) in enumerate(cases):
            new_c = self.visit_iHdlExpr(c)
            new_stm = self.visit_iHdlStatement(stm)
            cases[i] = (new_c, new_stm)

        if o.default is not None:
            o.default = self.visit_iHdlStatement(o.default)
        return o

    def visit_HdlStmWait(self, o):
        """
        :type o: HdlStmWait
        """
        self.visit_doc(o)
        o.val = self.visit_iHdlExpr(o.val)
        return o

    def visit_HdlStmIf(self, o):
        """
        :type o: HdlStmIf
        """
        self.visit_doc(o)
        o.cond = self.visit_iHdlExpr(o.cond)
        if o.if_true is not None:
            o.if_true = self.visit_iHdlStatement(o.if_true)

        for i, (c, stm) in enumerate(o.elifs):
            new_c = self.visit_iHdlExpr(c)
            new_stm = self.visit_iHdlStatement(stm)
            o.elifs[i] = (new_c, new_stm)

        if o.if_false is not None:
            o.if_false = self.visit_iHdlStatement(o.if_false)
        return o

    def visit_HdlStmFor(self, o):
        """
        :type o: HdlStmFor
        """
        self.visit_doc(o)
        o.init = self.visit_iHdlStatement(o.init)
        o.cond = self.visit_iHdlExpr(o.cond)
        o.step = self.visit_iHdlStatement(o.step)
        o.body = self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmForIn(self, o):
        """
        :type o: HdlStmForIn
        """
        self.visit_doc(o)
        self.visit_iHdlObj_list(o.var_defs, self.visit_main_obj)
        o.collection = self.visit_iHdlExpr(o.collection)
        o.body = self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmWhile(self, o):
        """
        :type o: HdlStmWhile
        """
        self.visit_doc(o)
        o.cond = self.visit_iHdlExpr(o.cond)
        o.body = self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmRepeat(self, o):
        """
        :type o: HdlStmRepeat
        """
        self.visit_doc(o)
        o.n = self.visit_iHdlExpr(o.n)
        o.body = self.visit_iHdlStatement(o.body)
        return o

    def visit_HdlStmReturn(self, o):
        """
        :type o: HdlStmReturn
        """
        self.visit_doc(o)
        if o.val is not None:
            o.val = self.visit_iHdlExpr(o.val)
        return o

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        """
        self.visit_doc(o)
        o.src = self.visit_iHdlExpr(o.src)
        o.dst = self.visit_iHdlExpr(o.dst)
        if o.event_delay is not None:
            o.event_delay = self.visit_iHdlExpr(o.event_delay)
        if o.time_delay is not None:
            o.time_delay = self.visit_iHdlExpr(o.time_delay)
        return o



from hdlConvertorAst.to.basic_hdl_sim_model.utils import verilog_slice_to_width
from hdlConvertorAst.hdlAst import HdlValueInt, HdlOpType, HdlOp, HdlValueId
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword


def BitsT(width, is_signed=False, bits_cls_name="Bits"):
    """
    Create an AST expression of Bits type constructor
    (reg/std_logic_vector equivalent for hwt)

    :type width: iHdlExpr
    """
    width = verilog_slice_to_width(width)

    if isinstance(width, int):
        width = HdlValueInt(width, None, None)

    args = [
        HdlValueId(bits_cls_name, obj=LanguageKeyword()),
        width,
    ]
    if is_signed is not None:
        args.append(HdlValueInt(int(is_signed), None, None))

    return HdlOp(HdlOpType.CALL, args)



from typing import Optional
from hdlConvertorAst.py_ver_compatibility import is_str


class LanguageKeyword(object):
    """
    Base class for keywords of target language.
    Used to notify that the name represents a language keyword
    and thus shall not be modified.
    """
    pass


class NameOccupiedErr(Exception):
    """
    An exception which is rised if the symbol name is used by a different
    object in current scope
    """

    def __init__(self, nameScope, name, usedOn):
        super(NameOccupiedErr, self).__init__()
        self.nameScope = nameScope
        self.name = name
        self.usedOn = usedOn

    def __str__(self):
        path = [self.name, ]
        ns = self.nameScope
        while ns is not None:
            n = ns.name
            if n is None:
                n = "<DEFAULT>"
            path.append(n)
            ns = ns.parent
        return "<%s in %s, usedOn=%r" % (self.__class__.__name__, "/".join(reversed(path)), self.usedOn)


class ObjectForNameNotFound(KeyError):
    """
    And exception which is rised if the name is not registered at all.
    """
    pass


_INVALID = object()


class NameScope(dict):
    """
    Scope of used names in hdl (node of hierarchical symbol table).
    Used to find an object for a HdlValueId or HdlValueId for an object and
    to resolve collision between different object of a same name if required.

    Contains mapping {name: obj}

    if name is discovered in scope it is converted to name_id
    where id is sequential number for prefix name\_

    :ivar ~.level: describes how deeply nested is this NameScopeItem in
        name hierarchy
    :ivar ~.cntrsForPrefixNames: counters for prefix names (for each name which
        has to be renamed there is a counter which is used to find a non occupied
        name faster)
    :ivar ~.reversed: a reverse dict to this dict ({obj: name})
    :ivar ~.serializer_ctx: an object with metainformations private to a current
        serializer
    :ivar ~.children: a dictionary {object which caused a NameScope fork: new NameScope}
    """
    @classmethod
    def make_top(cls, ignorecase):
        """
        Syntax shugar, call constructor with argument prefilled for a top NameScope

        :type ignorecase: bool
        :return: NameScope
        """
        return cls(None, None, ignorecase)

    def __init__(self, parent, name, ignorecase, debug=False):
        """
        :type parent: Optional[NameScope]
        :param name: name of object which this namescope belongs to
        :type name: Optional[str]
        :note: parent=None, name=None for global namescope
        :param ignorecase: if True the name comparison does not
            care about lowercase/uppercase
        :param debug: If True name scopes are not required to have
            forward declarations and for any object some name
            is always resolved without raising definition errors
        """
        super(NameScope, self).__init__()
        self.parent = parent
        self.ignorecase = ignorecase

        # some names are specified just as prefix and serializer
        # should resolve correct name for object
        # this happens for most of generated objects
        self.cntrsForPrefixNames = {}
        self.reversed = {}
        self.serializer_ctx = None
        self.children = {}
        self.name = name
        self.debug = debug

    def update(self, other):
        for k, v in other.items():
            self.register_name(k, v)

    # @internal
    def __incrPrefixCntr(self, prefix, currentVal):
        """
        :return: str
        """
        # [TODO] check if new name is not defined in any direction (parent-children)
        currentVal += 1
        self.cntrsForPrefixNames[prefix] = currentVal
        usableName = prefix + str(currentVal)
        return usableName

    # @internal
    def register_name(self, name, obj):
        # search if name is already defined on me and parents
        actual = self
        o = None

        if self.ignorecase:
            _name = name.lower()
        else:
            _name = name

        while actual is not None:
            try:
                o = actual[_name]
            except KeyError:
                actual = actual.parent
                continue
            break

        if o is obj:
            pass
        elif o is None:
            # we can use use the name, because it is not used
            # in any parent and that means we are not redefinig the symbol
            self[_name] = obj
            self.reversed[obj] = _name
        else:
            raise NameOccupiedErr(self, name, o)

    def get_child(self, name):
        assert is_str(name), name
        if self.ignorecase:
            name = name.lower()
        return self.children[name]

    def level_push(self, name):
        assert is_str(name), name
        if self.ignorecase:
            name = name.lower()

        i = self.children.get(name, None)
        if i is not None:
            # there is already a child with such a name
            return i

        if name not in self:
            if self.debug:
                self[name] = object()
            else:
                raise AssertionError(
                    name, "name not registered for any object in this scope")
        i = self.__class__(self, name, self.ignorecase)
        self.children[name] = i
        return i

    def level_pop(self):
        return self.parent

    def checked_name(self, suggested_name, obj):
        """
        Get a non occupied name in current scope
        if name is occupied or name ends with _ the new
        name is generated.

        :return: str
        """
        assert is_str(suggested_name), suggested_name
        if not suggested_name.endswith("_"):
            try:
                self.register_name(suggested_name, obj)
                return suggested_name
            except NameOccupiedErr:
                suggested_name += "_"

        actual = self
        while actual is not None:
            try:
                cntrVal = actual.cntrsForPrefixNames[suggested_name]
                break
            except KeyError:
                actual = actual.parent

        if actual is not None:
            # some parrent of self already have such a prefix counter
            usableName = actual.__incrPrefixCntr(
                suggested_name, cntrVal)
        else:
            # parrents and self does not have such a prefix counter
            # delete potentially existing prefix counterrs from children
            # and add prefix counter to self
            cntrVal = self.__discard_prefix_cntrs_from_children(suggested_name)
            usableName = self.__incrPrefixCntr(
                suggested_name, cntrVal)
        # setup for me and propagate to children
        self.register_name(usableName, obj)
        return usableName

    def __discard_prefix_cntrs_from_children(self, prefix):
        """
        Discard all prefix counters from all childrens in order to prevent
        children from looping trought all occupied names.
        """
        cntr_val = self.cntrsForPrefixNames.pop(prefix, -1)
        for c in self.children.values():
            cntr_val = max(
                cntr_val, c.__discard_prefix_cntrs_from_children(prefix))
        return cntr_val

    def get_object_and_scope_by_name(self, name):
        assert is_str(name), name
        if self.ignorecase:
            name = name.lower()
        actual = self
        while actual is not None:
            o = actual.get(name, _INVALID)
            if o is not _INVALID:
                return (actual, o)
            else:
                actual = actual.parent
        raise KeyError(name)

    def get_object_name(self, obj):
        """
        :return: str
        """
        assert obj is not None
        actual = self
        while actual is not None:
            n = actual.reversed.get(obj, None)
            if n is not None:
                return n
            actual = actual.parent

        raise ObjectForNameNotFound(obj)


class WithNameScope():
    """
    A syntax shugar, context manager which temporarly swaps the name_scope property on object
    """

    def __init__(self, obj, name_scope):
        """
        :type obj: an object which does have a name_scope stored in name_scope property
        :type name_scope: NameScope
        """
        self.obj = obj
        self.name_scope = name_scope

    def __enter__(self):
        self.original_name_scope = self.obj.name_scope
        self.obj.name_scope = self.name_scope

    def __exit__(self, type, value, traceback):
        self.obj.name_scope = self.original_name_scope



from hdlConvertorAst.hdlAst._bases import iHdlStatement
from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._statements import HdlStmBlock, HdlStmCaseType
from hdlConvertorAst.py_ver_compatibility import method_as_function
from hdlConvertorAst.to.hdlUtils import Indent
from hdlConvertorAst.to.systemc.expr import ToSystemcExpr
from hdlConvertorAst.to.verilog.stm import ToVerilog2005Stm


class ToSystemcStm(ToSystemcExpr):

    def visit_iHdlStatement_in_statement(self, stm):
        return method_as_function(ToVerilog2005Stm.visit_iHdlStatement_in_statement)(self, stm)

    def visit_iHdlStatement(self, o):
        """
        :type o: iHdlStatement
        """
        if isinstance(o, HdlIdDef):
            return self.visit_HdlIdDef(o)
        else:
            return super(ToSystemcStm, self).visit_iHdlStatement(o)

    def visit_HdlStmBlock(self, o):
        """
        :type o: HdlStmBlock
        """
        self.visit_doc(o)
        w = self.out.write
        w("{")
        w("\n")
        with Indent(self.out):
            for s in o.body:
                need_semi = self.visit_iHdlStatement(s)
                if need_semi:
                    w(";\n")
                else:
                    w("\n")
        w("}")
        return False

    def visit_HdlStmIf(self, o):
        return method_as_function(ToVerilog2005Stm.visit_HdlStmIf)(self, o)

    def visit_HdlStmProcess(self, o):
        """
        :type o: HdlStmProcess
        """
        w = self.out.write
        self.visit_doc(o)
        w("void ")
        w(o.labels[0])
        if o.trigger_constrain is not None:
            raise NotImplementedError()
        if isinstance(o.body, HdlStmBlock):
            w("() ")
            self.visit_HdlStmBlock(o.body)
            w("\n")
        else:
            w("() {\n")
            with Indent(self.out):
                req_semi = self.visit_iHdlStatement(o.body)
                if req_semi:
                    w(";\n")
                else:
                    w("\n")
            w("}\n")

    def visit_HdlStmAssign(self, o):
        """
        :type o: HdlStmAssign
        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        self.visit_iHdlExpr(o.dst)
        w(" = ")
        self.visit_iHdlExpr(o.src)
        return True

    def visit_HdlStmCase(self, o):
        """
        :type o: HdlStmCase

        :return: True if requires ;\\n after end
        """
        self.visit_doc(o)
        w = self.out.write
        if o.type != HdlStmCaseType.CASE:
            raise NotImplementedError()
        if o.uniq_constrain is not None:
            raise NotImplementedError()
        w("switch(")
        self.visit_iHdlExpr(o.switch_on)
        w(") {\n")
        cases = o.cases
        for k, stms in cases:
            w("case ")
            self.visit_iHdlExpr(k)
            w(":")
            with Indent(self.out):
                need_semi = self.visit_iHdlStatement_in_statement(stms)
                if need_semi:
                    w(";\n")
                else:
                    w("\n")
        defal = o.default
        if defal is not None:
            w("default:")
            with Indent(self.out):
                need_semi = self.visit_iHdlStatement_in_statement(defal)
                if need_semi:
                    w(";\n")
                else:
                    w("\n")
        w("}")
        return False

    def visit_HdlStmReturn(self, o):
        return method_as_function(ToVerilog2005Stm.visit_HdlStmReturn)(self, o)

    def visit_HdlStmContinue(self, o):
        return method_as_function(ToVerilog2005Stm.visit_HdlStmContinue)(self, o)

    def visit_HdlStmBreak(self, o):
        return method_as_function(ToVerilog2005Stm.visit_HdlStmBreak)(self, o)

    def visit_HdlStmThrow(self, o):
        """
        :type o: HdlStmThrow
        """
        self.visit_doc(o)
        w = self.out.write
        w("throw")
        if o.val is not None:
            w(" ")
            self.visit_iHdlExpr(o.val)
        w(";\n")



from hdlConvertorAst.hdlAst._expr import HdlOp, HdlValueId, HdlOpType
from hdlConvertorAst.hdlAst._statements import HdlStmAssign, HdlStmProcess
from hdlConvertorAst.hdlAst._structural import HdlModuleDef
from hdlConvertorAst.hdlAst import HdlValueInt


def collect_hdl_ids(expr, res):
    """
    :type expr: iHdlExpr
    :type res: Set[HdlValueId]
    """
    if isinstance(expr, HdlOp):
        for o in expr.ops:
            collect_hdl_ids(o, res)
    elif isinstance(expr, HdlValueId):
        res.add(expr)
    elif isinstance(expr, (HdlValueInt, bool, int, str, float)):
        pass
    else:
        raise NotImplementedError(expr)


def collect_indexes(expr):
    """
    Collect indexes from expression with optional index operator

    :type expr: iHdlExpr
    """
    if isinstance(expr, HdlOp) and expr.fn == HdlOpType.INDEX:
        assert len(expr.ops) == 2, expr.ops
        for i in collect_indexes(expr.ops[0]):
            yield i
        yield expr.ops[1]


def wrap_module_statements_to_processes(context):
    """
    Wrap statements which are not in any process instance in HdlStmProcess instance

    :type context: HdlContext
    """
    for o in context.objs:
        if isinstance(o, HdlModuleDef):
            objs = o.objs
            for i, obj in enumerate(objs):
                if isinstance(obj, HdlStmAssign):
                    p = HdlStmProcess()
                    p.body = obj
                    p.sensitivity = set()
                    for _i in collect_indexes(obj.dst):
                        collect_hdl_ids(_i, p.sensitivity)
                    collect_hdl_ids(obj.src, p.sensitivity)
                    objs[i] = p



"""
Use declarations collected from resolve_names
and use them to associate HdlValueIds to a HdlObjects.
"""
from itertools import chain

from hdlConvertorAst.hdlAst import HdlOp, HdlOpType, HdlValueId, HdlAll,\
    HdlValueInt, HdlIdDef, HdlStmAssign, HdlModuleDec, HdlModuleDef,\
    HdlCompInst, iHdlStatement, HdlTypeAuto
from hdlConvertorAst.to.hdl_ast_visitor import HdlAstVisitor
from hdlConvertorAst.translate.common.discover_declarations import WithNameScope


class ResolveNames(HdlAstVisitor):

    def __init__(self, name_scope):
        """
        :type name_scope: NameScope
        """
        super(ResolveNames, self).__init__()
        self.name_scope = name_scope
        self.in_args = False

    def visit_HdlIdDef(self, o):
        """
        :type o: HdlIdDef
        """
        self.name_scope.register_name(o.name, o)

    def visit_HdlModuleDec(self, o):
        """
        :type o: HdlModuleDec
        """
        with WithNameScope(self, self.name_scope.level_push(o.name)):
            for p in chain(o.params, o.ports):
                self.visit_HdlIdDef(p)

            for o2 in o.objs:
                raise NotImplementedError(o2)

    def visit_HdlModuleDef(self, o):
        """
        :type o: HdlModuleDef
        """
        if o.dec is not None:
            self.visit_HdlModuleDec(o.dec)

        with WithNameScope(self, self.name_scope.get_child(o.module_name.val)):
            for o2 in o.objs:
                self.visit_iHdlObj(o2)

    def visit_port_param_map(self, mod_name_scope, pmap):
        """
        :type pmap: List[HdlOp]
        """
        for pm in pmap:
            assert isinstance(pm, HdlOp) and\
                pm.fn == HdlOpType.MAP_ASSOCIATION, pm
            mod_port, connected_sig = pm.ops
            assert isinstance(mod_port, HdlValueId), mod_port
            orig_name_scope = self.name_scope
            try:
                self.name_scope = mod_name_scope
                self.visit_iHdlExpr(mod_port)
            finally:
                self.name_scope = orig_name_scope

            self.visit_iHdlExpr(connected_sig)

    def visit_HdlCompInst(self, o):
        """
        :type o: HdlCompInst
        """
        if o.name is not None:
            o.name.obj = o
        ns = self.name_scope
        mod_name_scope, mod_def = ns.get_object_and_scope_by_name(
            o.module_name.val)
        mod_name_scope = mod_name_scope.get_child(o.module_name.val)
        o.module_name.obj = mod_def
        self.visit_port_param_map(mod_name_scope, o.param_map)
        self.visit_port_param_map(mod_name_scope, o.port_map)

    def visit_HdlFunctionDef(self, o):
        """
        :type o: HdlFunctionDef
        """
        with WithNameScope(self, self.name_scope.level_push(o.name)):
            for p in o.params:
                self.visit_HdlIdDef(p)

            if o.return_t is not None:
                self.visit_iHdlExpr(o.return_t)
            for o2 in o.body:
                self.visit_main_obj(o2)
        return o

    def visit_iHdlExpr(self, o):
        """
        :type o: iHdlExpr
        """
        if isinstance(o, HdlOp):
            if o.fn is HdlOpType.DOT:
                # resolve only top id in id.id.id....
                self.visit_iHdlExpr(o.ops[0])
                assert isinstance(o.ops[1], HdlValueId), o.ops[1]

            elif o.fn in (HdlOpType.CALL, HdlOpType.PARAMETRIZATION):
                # mark argument list so we know that we should not search for
                # keyword ids
                orig_in_args = self.in_args
                try:
                    self.visit_iHdlExpr(o.ops[0])
                    self.in_args = True
                    for o2 in o.ops[1:]:
                        self.visit_iHdlExpr(o2)
                finally:
                    self.in_args = orig_in_args

            elif self.in_args and o.fn == HdlOpType.MAP_ASSOCIATION:
                # function keyword args, do not search for argument object
                assert isinstance(o.ops[0], HdlValueId), o.ops[0]
                for o2 in o.ops[1:]:
                    self.visit_iHdlExpr(o2)

            else:
                for o2 in o.ops:
                    self.visit_iHdlExpr(o2)

        elif isinstance(o, HdlValueId):
            _, o.obj = self.name_scope.get_object_and_scope_by_name(o.val)

        elif o is None or o is HdlAll or isinstance(
                o,  (HdlValueInt, float, str)):
            pass

        elif isinstance(o, (list, tuple)):
            for o2 in o:
                self.visit_iHdlExpr(o2)
        elif o is HdlTypeAuto:
            pass
        else:
            raise NotImplementedError(o)



import json

from hdlConvertorAst import hdlAst
from hdlConvertorAst.hdlAst import HdlContext, CodePosition, HdlOpType,\
    HdlDirection, HdlStmBlockJoinType, HdlValueInt, HdlValueId, HdlStmCase, HdlStmCaseType,\
    NON_INSTANCIABLE_NODES, HdlStmProcessTriggerConstrain
from hdlConvertorAst.py_ver_compatibility import is_str


KNOWN_NODES = {
    cls_name: getattr(hdlAst, cls_name)
    for cls_name in dir(hdlAst) if cls_name.startswith("Hdl")
}
KNOWN_NODES.update({
    "dict": dict,
    "list": list,
    "tuple": tuple,
    "str": str,
})


def _parse_hdlConvertor_json(j):
    # handle primitive types
    if j is None:
        return j
    elif isinstance(j, float):
        return j
    elif is_str(j):
        return HdlValueId(j)
    elif isinstance(j, int):
        return HdlValueInt(j, None, None)
    elif isinstance(j, list):
        return [_parse_hdlConvertor_json(_j) for _j in j]

    # load a hdlAst object
    cls = j["__class__"]
    cls = KNOWN_NODES[cls]
    consumed = {"__class__", }
    if cls in NON_INSTANCIABLE_NODES:
        assert len(j) == 1
        return cls

    elif cls in (dict, list, tuple):
        _items = j["items"]
        items = [_parse_hdlConvertor_json(i) for i in _items]
        if cls is dict:
            return {k: v for k, v in items}
        elif cls is tuple:
            return tuple(items)
        else:
            return items

    elif cls is str:
        return j["val"]

    elif cls is HdlValueInt:
        return HdlValueInt(j["val"], j.get("bits", None), j.get("base", None))

    argc = cls.__init__.__code__.co_argcount
    if argc == 1:
        o = cls()
    else:
        # load argumets for __init__
        argv = []
        # 1st is self
        arg_names = cls.__init__.__code__.co_varnames[1:argc]
        for a in arg_names:
            v = j.get(a, None)
            if a == "fn":
                v = getattr(HdlOpType, v)
            else:
                v = _parse_hdlConvertor_json(v)
            argv.append(v)
            consumed.add(a)
        o = cls(*argv)

    not_consumed = set(j.keys()).difference(consumed)
    if not_consumed:
        # load rest of the properties which were not in __init__ params
        for k in not_consumed:
            v = j[k]
            # there are few cases where object class is not specified specified
            # explicitly we have to handle them first
            if k == "position":
                _v = CodePosition()
                (
                    _v.start_line,
                    _v.start_column,
                    _v.stop_line,
                    _v.stop_column
                ) = v
            elif k == "direction":
                if v is None:
                    _v = v
                else:
                    _v = getattr(HdlDirection, v)
            elif k == "join_t":
                _v = getattr(HdlStmBlockJoinType, v)
            elif k == "trigger_constrain":
                _v = getattr(HdlStmProcessTriggerConstrain, v)
            elif cls is HdlStmCase and k == "type":
                _v = getattr(HdlStmCaseType, v)
            else:
                _v = _parse_hdlConvertor_json(v)
            setattr(o, k, _v)

    return o


def parse_hdlConvertor_json(j):
    """
    Convert loaded JSON (structure composed of list, dict, str, int, float, None)

    :return: HdlContext
    """
    assert isinstance(j, list), j
    ctx = HdlContext()
    for jo in j:
        o = _parse_hdlConvertor_json(jo)
        ctx.objs.append(o)
    return ctx


def parse_hdlConvertor_json_str(j_str):
    j = json.loads(j_str)
    return parse_hdlConvertor_json(j)


def parse_hdlConvertor_json_file(file_name):
    with open(file_name) as f:
        j = json.load(f)
        return parse_hdlConvertor_json(j)


if __name__ == "__main__":
    import os
    f_name = os.path.join(os.path.dirname(__file__), "..",
                          "tests", "data", "dff_async_reset.Verilog.json")
    print(parse_hdlConvertor_json_file(f_name))



from hdlConvertorAst.hdlAst._statements import HdlStmProcess
from hdlConvertorAst.hdlAst._structural import HdlModuleDef


class AddUniqueLabelsToAllProcesses():

    def __init__(self, name_scope, stm_outputs):
        """
        :type name_scope: NameScope
        :type stm_outputs: Dict[iHdlStatement, ]
        """
        self.name_scope = name_scope
        self.stm_outputs = stm_outputs

    def visit_HdlModuleDef(self, m):
        """
        :type m: HdlModuleDef
        """
        for o in m.objs:
            if isinstance(o, HdlStmProcess):
                if not o.labels:
                    outputs = self.stm_outputs[o]
                    if outputs:
                        suggested_name = "assig_process_" + outputs[0].val
                    else:
                        suggested_name = "proc"

                    o.labels.append(self.name_scope.checked_name(suggested_name, o))

    def context(self, context):
        """
        :type context: HdlContext
        """
        for o in context.objs:
            if isinstance(o, HdlModuleDef):
                try:
                    self.name_scope = self.name_scope.get_child(o.dec.name)
                    self.visit_HdlModuleDef(o)
                finally:
                    self.name_scope = self.name_scope.level_pop()
